/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QDoubleSpinBox>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QSpacerItem>
#include <QTabWidget>
#include <QWidget>

#include "../../core/xmlfile.h"

#include "featurecirclewidget.h"
#include "featurecylinderwidget.h"
#include "featurelinewidget.h"
#include "featurewidget.h"

TFeatureWidget::TFeatureWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QLabel								*orientation_label;
	QLabel								*orientation_i_label;
	QLabel								*orientation_j_label;
	QLabel								*orientation_k_label;
	QLabel								*position_label;
	QLabel								*position_x_label;
	QLabel								*position_y_label;
	QLabel								*position_z_label;
	QSpacerItem							*widget_vspacer;

	widget_layout = new QGridLayout(this);
	widget_layout->setSpacing(4);
	widget_layout->setContentsMargins(4,4,4,4);

	d_feature_type_tab = new QTabWidget(this);
	widget_layout->addWidget(d_feature_type_tab,0,0,1,3);

	d_feature_circle_widget = new TFeatureCircleWidget();
	d_feature_type_tab->addTab(d_feature_circle_widget,QStringLiteral("Circle"));

	d_feature_cylinder_widget = new TFeatureCylinderWidget();
	d_feature_type_tab->addTab(d_feature_cylinder_widget,QStringLiteral("Cylinder"));

	d_feature_line_widget = new TFeatureLineWidget();
	d_feature_type_tab->addTab(d_feature_line_widget,QStringLiteral("Line"));

	position_label = new QLabel(this);
	widget_layout->addWidget(position_label,1,0,1,1);

	position_x_label = new QLabel(this);
	widget_layout->addWidget(position_x_label,1,1,1,1);

	d_position_x_edit = new QLineEdit(this);
	d_position_x_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_position_x_edit,1,2,1,1);

	position_y_label = new QLabel(this);
	widget_layout->addWidget(position_y_label,2,1,1,1);

	d_position_y_edit = new QLineEdit(this);
	d_position_y_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_position_y_edit,2,2,1,1);

	position_z_label = new QLabel(this);
	widget_layout->addWidget(position_z_label,3,1,1,1);

	d_position_z_edit = new QLineEdit(this);
	d_position_z_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_position_z_edit,3,2,1,1);

	orientation_label = new QLabel(this);
	widget_layout->addWidget(orientation_label,4,0,1,1);

	orientation_i_label = new QLabel(this);
	widget_layout->addWidget(orientation_i_label,4,1,1,1);

	d_orientation_i_edit = new QLineEdit(this);
	d_orientation_i_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_orientation_i_edit,4,2,1,1);

	orientation_j_label = new QLabel(this);
	widget_layout->addWidget(orientation_j_label,5,1,1,1);

	d_orientation_j_edit = new QLineEdit(this);
	d_orientation_j_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_orientation_j_edit,5,2,1,1);

	orientation_k_label = new QLabel(this);
	widget_layout->addWidget(orientation_k_label,6,1,1,1);

	d_orientation_k_edit = new QLineEdit(this);
	d_orientation_k_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_orientation_k_edit,6,2,1,1);
	
	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,7,0,1,3);

	// defaults
	d_position_x_edit->setText(QStringLiteral("0.0000"));
	d_position_y_edit->setText(QStringLiteral("0.0000"));
	d_position_z_edit->setText(QStringLiteral("0.0000"));
	d_orientation_i_edit->setText(QStringLiteral("0.000000000"));
	d_orientation_j_edit->setText(QStringLiteral("0.000000000"));
	d_orientation_k_edit->setText(QStringLiteral("1.000000000"));
	d_data_changed = false;
	
	d_feature_type_tab->setCurrentIndex(0);		// default is circle

	QWidget::setTabOrder(d_position_x_edit,d_position_y_edit);
	QWidget::setTabOrder(d_position_y_edit,d_position_z_edit);
	QWidget::setTabOrder(d_position_z_edit,d_orientation_i_edit);
	QWidget::setTabOrder(d_orientation_i_edit,d_orientation_j_edit);
	QWidget::setTabOrder(d_orientation_j_edit,d_orientation_k_edit);

	position_label->setText(QStringLiteral("Position"));
	position_x_label->setText(QStringLiteral("X:"));
	position_y_label->setText(QStringLiteral("Y:"));
	position_z_label->setText(QStringLiteral("Z:"));
	orientation_label->setText(QStringLiteral("Orientation"));
	orientation_i_label->setText(QStringLiteral("I:"));
	orientation_j_label->setText(QStringLiteral("J:"));
	orientation_k_label->setText(QStringLiteral("K:"));

	connect(d_position_x_edit,&QLineEdit::editingFinished,this,&TFeatureWidget::Feature_Modified);
	connect(d_position_y_edit,&QLineEdit::editingFinished,this,&TFeatureWidget::Feature_Modified);
	connect(d_position_z_edit,&QLineEdit::editingFinished,this,&TFeatureWidget::Feature_Modified);
	connect(d_orientation_i_edit,&QLineEdit::editingFinished,this,&TFeatureWidget::Feature_Modified);
	connect(d_orientation_j_edit,&QLineEdit::editingFinished,this,&TFeatureWidget::Feature_Modified);
	connect(d_orientation_k_edit,&QLineEdit::editingFinished,this,&TFeatureWidget::Feature_Modified);

	connect(d_feature_circle_widget,&TFeatureCircleWidget::Data_Changed,this,&TFeatureWidget::Feature_Modified);
	connect(d_feature_cylinder_widget,&TFeatureCylinderWidget::Data_Changed,this,&TFeatureWidget::Feature_Modified);
	connect(d_feature_line_widget,&TFeatureLineWidget::Data_Changed,this,&TFeatureWidget::Feature_Modified);
	connect(d_feature_type_tab,&QTabWidget::currentChanged,this,&TFeatureWidget::Feature_Updated);
}

TFeatureWidget::~TFeatureWidget(void)
{
}

bool TFeatureWidget::Save_Data(
	const QString						&data_path) const
{
	if(d_data_changed)
	{
		if(this->Save(data_path))
		{
			d_data_changed = false;
			return true;
		}
		
		return false;
	}
	
	return true;
}
TVector3 TFeatureWidget::Position(void) const
{
	TVector3							pos;
	
	pos.x = d_position_x_edit->text().toDouble();
	pos.y = d_position_y_edit->text().toDouble();
	pos.z = d_position_z_edit->text().toDouble();

	return pos;
}

TVector3 TFeatureWidget::Orientation(void) const
{
	TVector3							vec;
	static const double					ZERO_EPSILON(0.000001);
	
	vec.i = d_orientation_i_edit->text().toDouble();
	vec.j = d_orientation_j_edit->text().toDouble();
	vec.k = d_orientation_k_edit->text().toDouble();
	
	if(vec.Length() < ZERO_EPSILON)
	{
		vec.Set(0,0,1);
	}
	else
	{
		vec.Normal();
	}

	return vec;
}

TFeatureWidget::TFeatureType TFeatureWidget::Active_Feature_Type(void) const
{
	QWidget								*active_widget;
	
	active_widget = d_feature_type_tab->currentWidget();
	
	if(active_widget == d_feature_cylinder_widget)
	{
		return TFeatureWidget::FEATURE_CYLINDER;
	}
	else if(active_widget == d_feature_line_widget)
	{
		return TFeatureWidget::FEATURE_LINE;
	}
	
	return TFeatureWidget::FEATURE_CIRCLE;
}

double TFeatureWidget::Feature_Circle_Diameter(void) const
{
	return d_feature_circle_widget->Diameter();
}

double TFeatureWidget::Feature_Cylinder_Diameter(void) const
{
	return d_feature_cylinder_widget->Diameter();
}

double TFeatureWidget::Feature_Cylinder_Length(void) const
{
	return d_feature_cylinder_widget->Length();
}

int TFeatureWidget::Feature_Cylinder_Levels(void) const
{
	return d_feature_cylinder_widget->Levels();
}

double TFeatureWidget::Feature_Line_Length(void) const
{
	return d_feature_line_widget->Length();
}

bool TFeatureWidget::Load_Data(
	const QString						&data_path)
{
	TXmlFile							xmlfile;
	int									version;
		
	try
	{
		xmlfile.Set_File_Name(QString("%1/features.xml").arg(data_path));
		xmlfile.Open("Features",&version);
		
		switch(version)
		{
			case 1:
				this->Load_Version1(&xmlfile);
				break;
		}
		
		d_data_changed = false;
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Loading features exception: %1").arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

void TFeatureWidget::Set_Position(
	const TVector3 						&pos)
{
	d_position_x_edit->setText(QString("%1").arg(pos.x,0,'f',4));
	d_position_y_edit->setText(QString("%1").arg(pos.y,0,'f',4));
	d_position_z_edit->setText(QString("%1").arg(pos.z,0,'f',4));
}

void TFeatureWidget::Set_Orientation(
	const TVector3 						&vec)
{
	static const double					ZERO_EPSILON(0.000001);
	
	if(vec.Length() < ZERO_EPSILON)
	{
		d_orientation_i_edit->setText(QStringLiteral("0.000000000"));
		d_orientation_j_edit->setText(QStringLiteral("0.000000000"));
		d_orientation_k_edit->setText(QStringLiteral("1.000000000"));
	}
	else
	{
		d_orientation_i_edit->setText(QString("%1").arg(vec.x,0,'f',9));
		d_orientation_j_edit->setText(QString("%1").arg(vec.y,0,'f',9));
		d_orientation_k_edit->setText(QString("%1").arg(vec.z,0,'f',9));
	}
}

void TFeatureWidget::Set_Active_Feature_Type(
	const TFeatureWidget::TFeatureType 	type)
{
	switch(type)
	{
		case TFeatureWidget::FEATURE_CIRCLE:
			d_feature_type_tab->setCurrentWidget(d_feature_circle_widget);
			break;
		
		case TFeatureWidget::FEATURE_CYLINDER:
			d_feature_type_tab->setCurrentWidget(d_feature_cylinder_widget);
			break;

		case TFeatureWidget::FEATURE_LINE:
			d_feature_type_tab->setCurrentWidget(d_feature_line_widget);
			break;
	}
}


void TFeatureWidget::Set_Feature_Circle_Diameter(
	const double 						&d)
{
	d_feature_circle_widget->Set_Diameter(d);
}

void TFeatureWidget::Set_Feature_Cylinder_Diameter(
	const double 						&d)
{
	d_feature_cylinder_widget->Set_Diameter(d);
}

void TFeatureWidget::Set_Feature_Cylinder_Length(
	const double 						&d)
{
	d_feature_cylinder_widget->Set_Length(d);
}

void TFeatureWidget::Set_Feature_Cylinder_Levels(
	const int 							i)
{
	d_feature_cylinder_widget->Set_Levels(i);
}

void TFeatureWidget::Set_Feature_Line_Length(
	const double 						&d)
{
	d_feature_line_widget->Set_Length(d);
}

void TFeatureWidget::Feature_Modified(void)
{
	d_data_changed = true;
	
	emit Feature_Updated();
}

bool TFeatureWidget::Save(
	const QString 						&data_path) const
{
	TXmlFile							xmlfile;
	QDomElement							en1;
	QDomElement							en2;
	TVector3							pt;

	try
	{
		xmlfile.Set_File_Name(QString("%1/features.xml").arg(data_path));
		xmlfile.Reset("Features",1);
		
		pt = this->Position();

		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Position"));
				
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("X"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(pt.x,0,'f',4));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Y"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(pt.y,0,'f',4));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Z"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(pt.z,0,'f',4));
		
		pt = this->Orientation();

		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Orientation"));
				
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("I"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(pt.x,0,'f',9));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("J"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(pt.y,0,'f',9));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("K"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(pt.z,0,'f',9));

		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Circle"));
				
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Diameter"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(this->Feature_Circle_Diameter(),0,'f',4));
		
		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Cylinder"));
				
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Diameter"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(this->Feature_Cylinder_Diameter(),0,'f',4));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Length"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(this->Feature_Cylinder_Length(),0,'f',4));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Levels"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(this->Feature_Cylinder_Levels()));
		
		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Line"));
				
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Length"));
		xmlfile.Write_Text_Node(&en2,QString("%1").arg(this->Feature_Line_Length(),0,'f',4));

		xmlfile.Write_File();
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Saving features exception: %1").arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

void TFeatureWidget::Load_Version1(
	TXmlFile							* const xml_file)
{
	QDomElement							en1;
	QDomElement							en2;
	TVector3							pt;
	
	assert(xml_file);
		
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Position"));
	
	en2 = xml_file->Get_Node(&en1,QStringLiteral("X"));
	pt.x = xml_file->Read_Text_Node(&en2).toDouble();
	
	en2 = xml_file->Get_Node(&en1,QStringLiteral("Y"));
	pt.y = xml_file->Read_Text_Node(&en2).toDouble();
	
	en2 = xml_file->Get_Node(&en1,QStringLiteral("Z"));
	pt.z = xml_file->Read_Text_Node(&en2).toDouble();
	
	this->Set_Position(pt);
	
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Orientation"));

	en2 = xml_file->Get_Node(&en1,QStringLiteral("I"));
	pt.x = xml_file->Read_Text_Node(&en2).toDouble();

	en2 = xml_file->Get_Node(&en1,QStringLiteral("J"));
	pt.y = xml_file->Read_Text_Node(&en2).toDouble();

	en2 = xml_file->Get_Node(&en1,QStringLiteral("K"));
	pt.z = xml_file->Read_Text_Node(&en2).toDouble();

	this->Set_Orientation(pt);
	
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Circle"));

	en2 = xml_file->Get_Node(&en1,QStringLiteral("Diameter"));
	this->Set_Feature_Circle_Diameter(xml_file->Read_Text_Node(&en2).toDouble());
	
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Cylinder"));

	en2 = xml_file->Get_Node(&en1,QStringLiteral("Diameter"));
	this->Set_Feature_Cylinder_Diameter(xml_file->Read_Text_Node(&en2).toDouble());

	en2 = xml_file->Get_Node(&en1,QStringLiteral("Length"));
	this->Set_Feature_Cylinder_Length(xml_file->Read_Text_Node(&en2).toDouble());

	en2 = xml_file->Get_Node(&en1,QStringLiteral("Levels"));
	this->Set_Feature_Cylinder_Levels(xml_file->Read_Text_Node(&en2).toInt());
	
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Line"));

	en2 = xml_file->Get_Node(&en1,QStringLiteral("Length"));
	this->Set_Feature_Line_Length(xml_file->Read_Text_Node(&en2).toDouble());
}

