/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QAction>
#include <QApplication>
#include <QScreen>
#include <QSettings>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QMenu>
#include <QMenuBar>
#include <QWidget>
#include <QDockWidget>
#include <QToolBar>
#include <QToolButton>
#include <QPixmap>
#include <QFile>
#include <QFileDialog>
#include <QCloseEvent>
#include <QByteArray>
#include <QTextStream>
#include <QTabWidget>
#include <QDir>
#include <QColor>
#include <QFont>
#include <algorithm>
#include <cmath>
#include <assert.h>

#include "../../core/messagebox.h"
#include "../../core/signalprocessing.h"
#include "../../core/graphwidget.h"
#include "../../core/vector3.h"
#include "../../core/referencesystem.h"
#include "../../core/interpolation.h"

#include "featurewidget.h"
#include "parametersettingswidget.h"
#include "conversionwidget.h"
#include "exportimportwidget.h"

#include "filterevaluation.h"

static bool DataPoint_Sort_Function(
	TVector2 							a,
	TVector2			 				b)
{
	return a.x < b.x;
}

static const double						PI2(6.283185307179586);
static const double						ZERO_EPSILON(0.00000001);
static const unsigned int				MINIMUM_SAMPLE_SIZE(64);
static const unsigned int				PADDING_SIZE((MINIMUM_SAMPLE_SIZE / 2) - 1);

TFilterEvaluation::TFilterEvaluation(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QAction								*action_help;
	QAction								*action_import_signal;
	QAction								*action_quit;
	QAction								*action_graph_fit;
	QToolBar							*toolbar;
	QWidget								*spacer_widget;
	QDockWidget							*feature_dock;
	QDockWidget							*parameters_dock;	
	QDockWidget							*conversion_dock;
	QDockWidget							*export_import_dock;
	QGridLayout							*widget_layout;
	QHBoxLayout							*copyright_hlayout;
	QLabel								*copyright_logo_label;
	QLabel								*copyright_text_label;
	QMenu								*file_menu;
	QMenu								*help_menu;
	QMenu								*view_menu;
	QMenuBar							*menu_bar;
	QWidget								*central_widget;
	QFont								copyright_font;
	QRect								desktop_rect;
	QRect								window_rect;
	QPoint								position;
	QSize								size;
	QDir								data_path;
	
	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);
	
	d_application_path = qApp->applicationDirPath();
	
#ifdef Q_OS_WIN
	d_data_path = d_application_path;
#else
	d_data_path = QDir::homePath();
#endif
	
	d_file_path = d_data_path;
	
	data_path.setPath(d_data_path);
	
	if(!data_path.exists(QStringLiteral(".filterevaluation")))
	{
		data_path.mkdir(QStringLiteral(".filterevaluation"));
	}
	
	data_path.cd(QStringLiteral(".filterevaluation"));
	d_data_path = data_path.absolutePath();
	
	d_settings = new QSettings(d_data_path + QStringLiteral("/filterevaluation_settings.ini"),QSettings::IniFormat,this);
	
	d_file_path = d_settings->value("File_Path", d_file_path).toString();

	position = d_settings->value("Mainwindow_Position", QPoint(30,30)).toPoint();
	size = d_settings->value("Mainwindow_Size", QSize(1001,601)).toSize();
	
	desktop_rect = QGuiApplication::primaryScreen()->availableGeometry();
	
	window_rect.moveTopLeft(position);
	window_rect.setSize(size);
	
	window_rect = window_rect & desktop_rect;
	
	if(window_rect.isValid() && window_rect.width() > 500 && window_rect.height() > 500)
	{
		this->resize(window_rect.size());
		this->move(window_rect.topLeft());
	}
	else
	{
		this->resize(1001,601);
		this->move(QPoint(30,30));
	}

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);
	
	widget_layout = new QGridLayout(central_widget);

	d_graph_tab = new QTabWidget(central_widget);
	widget_layout->addWidget(d_graph_tab,0,0,1,1);
	
	d_generated_views_tab = new QTabWidget();
	d_generated_views_tab->setTabPosition(QTabWidget::South);

	d_generated_signal_graph_widget = new TGraphWidget();
	d_generated_views_tab->addTab(d_generated_signal_graph_widget,QStringLiteral("Signal"));

	d_generated_wave_graph_widget = new TGraphWidget();
	d_generated_views_tab->addTab(d_generated_wave_graph_widget,QStringLiteral("Waveform"));

	d_graph_tab->addTab(d_generated_views_tab,QStringLiteral("Generated"));

	d_imported_views_tab = new QTabWidget();
	d_imported_views_tab->setTabPosition(QTabWidget::South);

	d_imported_signal_graph_widget = new TGraphWidget();
	d_imported_views_tab->addTab(d_imported_signal_graph_widget,QStringLiteral("Signal"));
	
	d_imported_wave_graph_widget = new TGraphWidget();
	d_imported_views_tab->addTab(d_imported_wave_graph_widget,QStringLiteral("Waveform"));

	d_graph_tab->addTab(d_imported_views_tab,QStringLiteral("Imported"));

	copyright_hlayout = new QHBoxLayout();
	
	copyright_text_label = new QLabel(central_widget);
	copyright_text_label->setFont(copyright_font);
	copyright_text_label->setAlignment(Qt::AlignBottom|Qt::AlignLeading|Qt::AlignLeft);
	copyright_hlayout->addWidget(copyright_text_label);

	copyright_logo_label = new QLabel(central_widget);
	copyright_logo_label->setMinimumSize(100,40);
	copyright_logo_label->setMaximumSize(100,40);
	copyright_hlayout->addWidget(copyright_logo_label);
	widget_layout->addLayout(copyright_hlayout,1,0,1,1);

	action_import_signal = new QAction(this);
	action_import_signal->setIcon(QIcon(":/toolbar/import_signal_x32.png"));

	d_action_export_signal = new QAction(this);
	d_action_export_signal->setIcon(QIcon(":/toolbar/export_signal_x32.png"));

	d_action_export_dmis = new QAction(this);
	d_action_export_dmis->setIcon(QIcon(":/toolbar/export_dmis_x32.png"));

	action_quit = new QAction(this);
	action_help = new QAction(this);
	
	d_action_graph_bar = new QAction(this);
	d_action_graph_bar->setIcon(QIcon(":/toolbar/bar_graph_x32.png"));

	d_action_graph_line = new QAction(this);
	d_action_graph_line->setIcon(QIcon(":/toolbar/line_graph_x32.png"));

	d_action_graph_point = new QAction(this);
	d_action_graph_point->setIcon(QIcon(":/toolbar/point_graph_x32.png"));
	
	action_graph_fit = new QAction(this);
	action_graph_fit->setIcon(QIcon(":/toolbar/scale_to_fit_x32.png"));

	menu_bar = new QMenuBar(this);
	this->setMenuBar(menu_bar);

	file_menu = new QMenu(menu_bar);
	menu_bar->addMenu(file_menu);
	file_menu->addAction(action_import_signal);
	file_menu->addSeparator();
	file_menu->addAction(d_action_export_signal);
	file_menu->addAction(d_action_export_dmis);
	file_menu->addSeparator();
	file_menu->addAction(action_quit);

	view_menu = new QMenu(menu_bar);
	menu_bar->addMenu(view_menu);

	help_menu = new QMenu(menu_bar);
	menu_bar->addMenu(help_menu);
	help_menu->addAction(action_help);

	menu_bar->addAction(file_menu->menuAction());
	menu_bar->addAction(view_menu->menuAction());
	menu_bar->addAction(help_menu->menuAction());
	
	toolbar = new QToolBar(this);
	toolbar->setIconSize(QSize(32,32));
	this->addToolBar(Qt::TopToolBarArea,toolbar);

	toolbar->addAction(action_import_signal);
	toolbar->addSeparator();
	toolbar->addAction(d_action_export_signal);
	toolbar->addAction(d_action_export_dmis);

	spacer_widget = new QWidget();
	spacer_widget->setSizePolicy(QSizePolicy::MinimumExpanding,QSizePolicy::Minimum);
	toolbar->addWidget(spacer_widget);

	toolbar->addAction(d_action_graph_bar);
	toolbar->addAction(d_action_graph_line);
	toolbar->addAction(d_action_graph_point);
	toolbar->addAction(action_graph_fit);

	feature_dock = new QDockWidget(this);
	feature_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,feature_dock);
	
	d_feature_widget = new TFeatureWidget();
	feature_dock->setWidget(d_feature_widget);
	
	export_import_dock = new QDockWidget(this);
	export_import_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,export_import_dock);

	d_export_import_widget = new TExportImportWidget();
	export_import_dock->setWidget(d_export_import_widget);

	parameters_dock = new QDockWidget(this);
	parameters_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::RightDockWidgetArea,parameters_dock);
	
	d_parameters_settings_widget = new TParameterSettingsWidget();
	parameters_dock->setWidget(d_parameters_settings_widget);
	
	conversion_dock = new QDockWidget(this);
	conversion_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::RightDockWidgetArea,conversion_dock);

	d_conversion_widget = new TConversionWidget();
	conversion_dock->setWidget(d_conversion_widget);

	view_menu->addAction(feature_dock->toggleViewAction());
	view_menu->addAction(parameters_dock->toggleViewAction());
	view_menu->addAction(conversion_dock->toggleViewAction());
	view_menu->addAction(export_import_dock->toggleViewAction());
	
	d_msg_box = new TMessageBox(this);

	// defaults
	d_generated_signal_graph_widget->SetHLegendMode(TGraphWidget::LOG10_LEGEND_PRECISION);
	d_generated_signal_graph_widget->SetGraphType(TGraphWidget::BAR_GRAPH);
	d_generated_signal_graph_widget->EnableYIncludesZero(true);
	d_generated_signal_graph_widget->HorizontalLegend(QStringLiteral("Undulations Per Unit"));

	d_generated_wave_graph_widget->SetGraphType(TGraphWidget::LINE_GRAPH);
	d_generated_wave_graph_widget->HorizontalLegend(QStringLiteral("Position"));

	d_imported_signal_graph_widget->SetHLegendMode(TGraphWidget::LOG10_LEGEND_PRECISION);
	d_imported_signal_graph_widget->SetVLegendMode(TGraphWidget::INTEGER_LEGEND_PRECISION);
	d_imported_signal_graph_widget->SetGraphType(TGraphWidget::BAR_GRAPH);
	d_imported_signal_graph_widget->EnableYIncludesZero(true);
	d_imported_signal_graph_widget->SetScaleMinimumY(110,false);
	d_imported_signal_graph_widget->HorizontalLegend(QStringLiteral("Undulations Per Unit"));
	
	d_imported_wave_graph_widget->SetGraphType(TGraphWidget::LINE_GRAPH);
	d_imported_wave_graph_widget->HorizontalLegend(QStringLiteral("Position"));

	d_parameters_settings_widget->Load_Data(d_data_path);
	d_feature_widget->Load_Data(d_data_path);
	d_feature_widget->Set_Active_Feature_Type(static_cast<TFeatureWidget::TFeatureType>(d_settings->value("Active_Feature_Type", 0).toInt()));
	
	this->Update_Signal();	

	copyright_logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));

	feature_dock->setWindowTitle(QStringLiteral("Feature"));
	parameters_dock->setWindowTitle(QStringLiteral("Parameters"));
	conversion_dock->setWindowTitle(QStringLiteral("Conversion"));
	export_import_dock->setWindowTitle(QStringLiteral("Waveform"));

	this->setWindowTitle(QStringLiteral("Filter Evaluation"));
	
	d_generated_signal_graph_widget->Title(QStringLiteral("Frequencies"));
	d_generated_signal_graph_widget->VerticalLegend(QStringLiteral("Amplitude"));
	
	d_generated_wave_graph_widget->Title(QStringLiteral("Waveform"));
	d_generated_wave_graph_widget->VerticalLegend(QStringLiteral("Amplitude"));
	
	d_imported_signal_graph_widget->Title(QStringLiteral("Frequencies"));
	d_imported_signal_graph_widget->VerticalLegend(QStringLiteral("Relative Amplitude %"));
	
	d_imported_wave_graph_widget->Title(QStringLiteral("Waveform"));
	d_imported_wave_graph_widget->VerticalLegend(QStringLiteral("Amplitude"));

	copyright_text_label->setText(QStringLiteral("Copyright @2024 Select Calibration Incorporated"));
	action_import_signal->setText(QStringLiteral("Import Signal"));
	d_action_export_signal->setText(QStringLiteral("Export Signal"));
	d_action_export_dmis->setText(QStringLiteral("Export DMIS"));
	action_quit->setText(QStringLiteral("Quit"));
	action_help->setText(QStringLiteral("About"));
	action_graph_fit->setText(QStringLiteral("Fit"));
	file_menu->setTitle(QStringLiteral("File"));
	view_menu->setTitle(QStringLiteral("View"));
	help_menu->setTitle(QStringLiteral("Help"));
		
	d_action_graph_bar->setToolTip(QStringLiteral("Bar Graph"));
	d_action_graph_line->setToolTip(QStringLiteral("Line Graph"));
	d_action_graph_point->setToolTip(QStringLiteral("Point Graph"));
	action_graph_fit->setToolTip(QStringLiteral("Scale To Fit"));
	action_import_signal->setToolTip(QStringLiteral("Import Signal"));
	d_action_export_signal->setToolTip(QStringLiteral("Export Signal"));
	d_action_export_dmis->setToolTip(QStringLiteral("Export DMIS"));

	connect(d_action_graph_bar,&QAction::triggered,this,&TFilterEvaluation::Graph_Bar);
	connect(d_action_graph_line,&QAction::triggered,this,&TFilterEvaluation::Graph_Line);
	connect(d_action_graph_point,&QAction::triggered,this,&TFilterEvaluation::Graph_Point);
	connect(action_graph_fit,&QAction::triggered,this,&TFilterEvaluation::Graph_Scale_To_Fit);
	connect(action_import_signal,&QAction::triggered,this,&TFilterEvaluation::Import_Signal_Data);
	connect(d_action_export_signal,&QAction::triggered,this,&TFilterEvaluation::Write_Signal);
	connect(d_action_export_dmis,&QAction::triggered,this,&TFilterEvaluation::Write_DMIS);
	connect(action_quit,&QAction::triggered,this,&TFilterEvaluation::close);
	connect(action_help,&QAction::triggered,this,&TFilterEvaluation::Help_About);
	connect(d_graph_tab,&QTabWidget::currentChanged,this,&TFilterEvaluation::View_Changed);
	connect(d_generated_views_tab,&QTabWidget::currentChanged,this,&TFilterEvaluation::Sub_View_Changed);
	connect(d_imported_views_tab,&QTabWidget::currentChanged,this,&TFilterEvaluation::Sub_View_Changed);

	connect(d_parameters_settings_widget,&TParameterSettingsWidget::Parameters_Changed,this,&TFilterEvaluation::Update_Signal);
	connect(d_feature_widget,&TFeatureWidget::Feature_Updated,this,&TFilterEvaluation::Update_Signal);
}

TFilterEvaluation::~TFilterEvaluation(void)
{
}

void TFilterEvaluation::Graph_Bar(void)
{
	d_generated_signal_graph_widget->SetGraphType(TGraphWidget::BAR_GRAPH);
	d_generated_signal_graph_widget->UpdateGraph();
	
	d_imported_signal_graph_widget->SetGraphType(TGraphWidget::BAR_GRAPH);
	d_imported_signal_graph_widget->UpdateGraph();
}

void TFilterEvaluation::Graph_Line(void)
{
	d_generated_signal_graph_widget->SetGraphType(TGraphWidget::LINE_GRAPH);
	d_generated_signal_graph_widget->UpdateGraph();
	
	d_imported_signal_graph_widget->SetGraphType(TGraphWidget::LINE_GRAPH);
	d_imported_signal_graph_widget->UpdateGraph();
}

void TFilterEvaluation::Graph_Point(void)
{
	d_generated_signal_graph_widget->SetGraphType(TGraphWidget::POINT_GRAPH);
	d_generated_signal_graph_widget->UpdateGraph();

	d_imported_signal_graph_widget->SetGraphType(TGraphWidget::POINT_GRAPH);
	d_imported_signal_graph_widget->UpdateGraph();
}

void TFilterEvaluation::Graph_Scale_To_Fit(void)
{
	d_generated_signal_graph_widget->ScaleToFit();
	d_imported_signal_graph_widget->ScaleToFit();
	d_generated_wave_graph_widget->ScaleToFit();
	d_imported_wave_graph_widget->ScaleToFit();
}

void TFilterEvaluation::Update_Signal(void)
{
	TFourierTransform					fourier_transform;
	std::vector<TVector2>				fourier_data;
	std::vector<TVector2>				fourier_signal_data;
	std::vector<TParameterSettingsWidget::TEntryItem> entry_items;
	std::vector<TParameterSettingsWidget::TEntryItem>::const_iterator entry_iter;
	std::vector<TVector2>::iterator		iter;
	TFeatureWidget::TFeatureType		feature_type;
	TVector2							pnt;
	double								data_pos;
	double								max_value;
	double								data_pos_increment;
	double								modulation_multiplier;
	double								length(0.0);
	double								dval;
	int									set;
	int									cntr;
	int									sample_size;
	static const double					CONVERT_RAD_TO_DEG(57.2957795131);
	static const double					CONVERT_DEG_TO_RAD(0.017453292519943);

	d_graph_tab->setCurrentWidget(d_generated_views_tab);
	
	feature_type = d_feature_widget->Active_Feature_Type();
		
	d_generated_signal_graph_widget->DeleteDataSets();
	d_generated_wave_graph_widget->DeleteDataSets();

	switch(feature_type)
	{
		case TFeatureWidget::FEATURE_CIRCLE:
			d_generated_signal_graph_widget->HorizontalLegend(QStringLiteral("Undulations Per Revolution"));
			d_generated_wave_graph_widget->HorizontalLegend(QStringLiteral("Angle in Degrees"));
			
			d_conversion_widget->Set_Circular_Diameter(d_feature_widget->Feature_Circle_Diameter());
			break;
			
		case TFeatureWidget::FEATURE_CYLINDER:
			d_generated_signal_graph_widget->HorizontalLegend(QStringLiteral("Undulations Per Revolution"));
			d_generated_wave_graph_widget->HorizontalLegend(QStringLiteral("Angle in Degrees"));

			d_conversion_widget->Set_Circular_Diameter(d_feature_widget->Feature_Cylinder_Diameter());
			break;
			
		case TFeatureWidget::FEATURE_LINE:
			length = d_feature_widget->Feature_Line_Length();
			d_generated_signal_graph_widget->HorizontalLegend(QStringLiteral("Undulations Per Unit Length"));
			d_generated_wave_graph_widget->HorizontalLegend(QStringLiteral("Length in mm"));
			break;
	}
	
	d_conversion_widget->Set_Linear_Length(d_feature_widget->Feature_Line_Length());
	
	entry_items = d_parameters_settings_widget->Entry_Items();

	d_action_export_signal->setEnabled(false);
	d_action_export_dmis->setEnabled(false);

	if(!entry_items.size())
	{
		d_generated_signal_graph_widget->UpdateGraph();
		d_generated_wave_graph_widget->UpdateGraph();
		return;
	}
	
	d_action_export_signal->setEnabled(true);
	d_action_export_dmis->setEnabled(true);

	sample_size = this->Get_Minimum_Sample_Size(d_parameters_settings_widget->Maximum_Undulations());
	
	d_export_import_widget->Set_Export_Size(sample_size);
	sample_size += (PADDING_SIZE * 2);
	
	d_signal_data.resize(sample_size);
		
	modulation_multiplier = static_cast<double>(d_parameters_settings_widget->Modulation());
	modulation_multiplier /= 281.135761868635669;	// +/-1 deg modulation with an input of 100%
	
	data_pos_increment = PI2 / static_cast<double>(sample_size - PADDING_SIZE * 2);

	data_pos = 0.0 - (static_cast<double>(PADDING_SIZE) * data_pos_increment);
	for(cntr = 0;cntr < sample_size;++cntr,data_pos += data_pos_increment)
	{
		dval = data_pos + sin(data_pos) * modulation_multiplier;
		d_signal_data[cntr].Set(dval,0.0);
	}
	
	for(entry_iter = entry_items.begin();entry_iter != entry_items.end();++entry_iter)
	{
		this->Add_Signal_Entry((*entry_iter).undulations,(*entry_iter).phase * CONVERT_DEG_TO_RAD,(*entry_iter).amplitude);
	}

	fourier_signal_data = std::vector<TVector2>(d_signal_data.begin() + PADDING_SIZE,d_signal_data.end() - PADDING_SIZE);
	fourier_signal_data = this->Normalize_Signal_Data(fourier_signal_data);
	fourier_data = fourier_transform.Process(fourier_signal_data);
		
	max_value = this->Get_Max_Value(fourier_data);
	this->Normalize(&fourier_data,max_value,d_parameters_settings_widget->Maximum_Amplitude());
	this->Convert_To_Undulations(&fourier_data);

	if(fourier_data.size())
	{
		fourier_data.resize(fourier_data.size() / 4);	// show only relevant parts
	}

	set = d_generated_signal_graph_widget->CreateDataSet(QColor(160,0,0));
	d_generated_signal_graph_widget->AddPoints(set,fourier_data);
	
	set = d_generated_wave_graph_widget->CreateDataSet(QColor(160,0,0));

	if(feature_type == TFeatureWidget::FEATURE_LINE)
	{
		dval = length / PI2;		
	}
	
	// add end point
	pnt = fourier_signal_data[0];
	pnt.x = PI2;
		
	fourier_signal_data.push_back(pnt);

	
	for(iter = fourier_signal_data.begin();iter != fourier_signal_data.end();++iter)
	{
		if(feature_type == TFeatureWidget::FEATURE_LINE)
		{
			pnt.Set((*iter).x * dval,(*iter).y);
		}
		else
		{
			pnt.Set((*iter).x * CONVERT_RAD_TO_DEG,(*iter).y);
		}
		
		d_generated_wave_graph_widget->AddPoint(set,pnt);
	}

	d_generated_signal_graph_widget->UpdateGraph();
	d_generated_wave_graph_widget->UpdateGraph();
}

void TFilterEvaluation::Write_Signal(void)
{
	TFeatureWidget::TFeatureType		feature_type;
	QString								file_name;
	int									index;
	
	if(!d_signal_data.size())
	{
		return;
	}
	
	d_graph_tab->setCurrentWidget(d_generated_signal_graph_widget);
	
	file_name = QFileDialog::getSaveFileName(this,QStringLiteral("Generate Signal File"),d_file_path,QStringLiteral("Signal File (*.signal)"));
	
	if(!file_name.length())
	{
		return;
	}
	
	index = file_name.lastIndexOf('/');
	
	d_file_path = file_name;
	
	if(!(index < 0))
	{
		d_file_path.truncate(index);
	}
	
	feature_type = d_feature_widget->Active_Feature_Type();
				
	switch(feature_type)
	{
		case TFeatureWidget::FEATURE_CIRCLE:
		case TFeatureWidget::FEATURE_CYLINDER:
			this->Write_Signal_Circle(file_name);
			break;
			
		case TFeatureWidget::FEATURE_LINE:
			this->Write_Signal_Line(file_name);
			break;
	}
}

void TFilterEvaluation::Write_DMIS(void)
{
	TFeatureWidget::TFeatureType		feature_type;
	QString								file_name;
	int									index;
	
	if(!d_signal_data.size())
	{
		return;
	}

	d_graph_tab->setCurrentWidget(d_generated_signal_graph_widget);
	
	file_name = QFileDialog::getSaveFileName(this,QStringLiteral("Generate DMIS Program"),d_file_path,QStringLiteral("DMIS File (*.dmi *.dmis)"));
	
	if(!file_name.length())
	{
		return;
	}
	
	index = file_name.lastIndexOf('/');
	
	d_file_path = file_name;
	
	if(!(index < 0))
	{
		d_file_path.truncate(index);
	}
	
	feature_type = d_feature_widget->Active_Feature_Type();
				
	switch(feature_type)
	{
		case TFeatureWidget::FEATURE_CIRCLE:
			this->Write_DMIS_Circle(file_name);
			break;
			
		case TFeatureWidget::FEATURE_CYLINDER:
			this->Write_DMIS_Cylinder(file_name);
			break;
			
		case TFeatureWidget::FEATURE_LINE:
			this->Write_DMIS_Line(file_name);
			break;
	}
}

void TFilterEvaluation::Import_Signal_Data(void)
{
	TFourierTransform					fourier_transform;
	TFilterEvaluation::TSignalType		signal_type;
	QString								file_name;
	std::vector<TVector2>				signal_data;
	std::vector<TVector2>				fourier_data;
	std::vector<TVector2>				fourier_signal_data;
	std::vector<TVector2>::iterator		iter;
	TVector2							pnt;
	double								max_value;
	double								length;
	double								dval(1.0);
	int									set;
	int									index;
	bool								valid;
	static const double					CONVERT_RAD_TO_DEG(57.2957795131);
	
	file_name = QFileDialog::getOpenFileName(this,QStringLiteral("Select Input Signal File"),d_file_path,QStringLiteral("Signal File (*.signal)"));
	
	if(!file_name.length())
	{
		return;
	}
		
	index = file_name.lastIndexOf('/');
	
	d_file_path = file_name;
	
	if(!(index < 0))
	{
		d_file_path.truncate(index);
	}
	
	signal_data = this->Read_Signal_File(file_name,&valid,&signal_type,&length);
		
	if(!valid)
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Signal data file import error."));
		d_msg_box->setDetailedText(d_last_error);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
		
		return;
	}
		
	if(signal_data.size())
	{
		d_imported_signal_graph_widget->DeleteDataSets();
		d_imported_wave_graph_widget->DeleteDataSets();

		if(signal_type == TFilterEvaluation::CIRCLE_SIGNAL)
		{
			d_imported_signal_graph_widget->HorizontalLegend(QStringLiteral("Undulations Per Revolution"));
			d_imported_wave_graph_widget->HorizontalLegend(QStringLiteral("Angle in Degrees"));
		}
		else if(signal_type == TFilterEvaluation::LINE_SIGNAL)
		{
			d_imported_signal_graph_widget->HorizontalLegend(QStringLiteral("Undulations Per Unit Length"));
			d_imported_wave_graph_widget->HorizontalLegend(QStringLiteral("Length in mm"));
		}
		else
		{
			d_msg_box->setText(QStringLiteral("Error"));
			d_msg_box->setInformativeText(QStringLiteral("Signal type not recognized."));
			d_msg_box->setDetailedText(QString());
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			d_msg_box->setIcon(QMessageBox::Warning);
			
			d_msg_box->exec();
			
			return;
		}
		
		d_export_import_widget->Set_Import_Size(static_cast<int>(signal_data.size()));


		fourier_signal_data = this->Normalize_Signal_Data(signal_data);
		fourier_data = fourier_transform.Process(fourier_signal_data);
			
		max_value = this->Get_Max_Value(fourier_data);
		
		this->Normalize(&fourier_data,max_value,100.0);
		this->Convert_To_Undulations(&fourier_data);

		if(fourier_data.size())
		{
			fourier_data.resize(fourier_data.size() / 4);	// show only relevant parts
		}

		set = d_imported_signal_graph_widget->CreateDataSet(QColor(0,0,160));
		d_imported_signal_graph_widget->AddPoints(set,fourier_data);

		set = d_imported_wave_graph_widget->CreateDataSet(QColor(0,0,160));

		if(signal_type == TFilterEvaluation::LINE_SIGNAL)
		{
			dval = length / 6.28318530718;
		}
		
		// add end point
		pnt = signal_data[0];
		pnt.x = PI2;
		
		signal_data.push_back(pnt);

		for(iter = signal_data.begin();iter != signal_data.end();++iter)
		{
			if(signal_type == TFilterEvaluation::LINE_SIGNAL)
			{
				pnt.Set((*iter).x * dval,(*iter).y);
			}
			else
			{
				pnt.Set((*iter).x * CONVERT_RAD_TO_DEG,(*iter).y);
			}
			
			d_imported_wave_graph_widget->AddPoint(set,pnt);
		}
		
		d_imported_signal_graph_widget->UpdateGraph();
		d_imported_wave_graph_widget->UpdateGraph();
		
		d_graph_tab->setCurrentWidget(d_imported_views_tab);
	}
}

void TFilterEvaluation::Help_About(void)
{
	d_msg_box->setText(QStringLiteral("Filter Evaluation"));
	d_msg_box->setInformativeText(QStringLiteral("Filter Evaluation - Version 1.0"));
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Ok);
	d_msg_box->setDefaultButton(QMessageBox::Ok);
	d_msg_box->setIcon(QMessageBox::Information);
	
	d_msg_box->exec();
}

void TFilterEvaluation::View_Changed(
	int									index)
{
	int									sub_index;
	
	switch(index)
	{
		case 0:	// generated
			
			sub_index = d_generated_views_tab->currentIndex();
			
			d_action_graph_bar->setEnabled(sub_index == 0);
			d_action_graph_line->setEnabled(sub_index == 0);
			d_action_graph_point->setEnabled(sub_index == 0);

			break;
			
		case 1:	// imported
			
			sub_index = d_imported_views_tab->currentIndex();
			
			d_action_graph_bar->setEnabled(sub_index == 0);
			d_action_graph_line->setEnabled(sub_index == 0);
			d_action_graph_point->setEnabled(sub_index == 0);

			break;
	}
}

void TFilterEvaluation::Sub_View_Changed(
	int									index)
{	
	d_action_graph_bar->setEnabled(index == 0);
	d_action_graph_line->setEnabled(index == 0);
	d_action_graph_point->setEnabled(index == 0);
}

void TFilterEvaluation::closeEvent(
	QCloseEvent							*event)
{
	d_parameters_settings_widget->Save_Data(d_data_path);
	d_feature_widget->Save_Data(d_data_path);
	
	d_settings->setValue("Mainwindow_Position", this->pos());
	d_settings->setValue("Mainwindow_Size", this->size());
	d_settings->setValue("File_Path", d_file_path);
	d_settings->setValue("Active_Feature_Type", static_cast<int>(d_feature_widget->Active_Feature_Type()));
	
	event->accept();
}

double TFilterEvaluation::Get_Max_Value(
	const std::vector<TVector2>			&data) const
{
	bool								init(true);
	double								result(0.0);
	std::vector<TVector2>::const_iterator iter;
	
	for(iter = data.begin();iter != data.end();++iter)
	{
		if(init)
		{
			init = false;
			result = (*iter).y;
		}
		else
		{
			if((*iter).y > result)
			{
				result = (*iter).y;
			}
		}
	}
	
	return result;
}

void TFilterEvaluation::Normalize(
	std::vector<TVector2>				* const data,
	const double						&max,
	const double						&range) const
{
	std::vector<TVector2>::iterator		iter;
	
	assert(data);
	
	if(max < ZERO_EPSILON)
	{
		return;
	}
	
	if(range < ZERO_EPSILON)
	{
		return;
	}

	for(iter = data->begin();iter != data->end();++iter)
	{
		(*iter).y /= max;
		(*iter).y *= range;
	}
}

std::vector<TVector2> TFilterEvaluation::Normalize_Signal_Data(
	const std::vector<TVector2> 		&signal_data) const
{
	std::vector<TVector2>				normalized_signal;
	std::vector<TVector2>				signal_data_pass0;
	std::vector<TVector2>				signal_data_pass1;
	std::vector<TVector2>				signal_data_pass2;
	std::vector<TVector2>::const_iterator iter;
	TVector2							curve_data[3];
	TVector2							first_point;
	TVector2							second_point;
	TVector2							last_second_point;
	TVector2							pnt;
	double								dval;
	double								angle;
	double								angle_increment;
	bool								init;

	if(signal_data.size() > 2)
	{
		signal_data_pass0 = signal_data;
		
		std::sort(signal_data_pass0.begin(), signal_data_pass0.end(), DataPoint_Sort_Function);
		
		// first iteration
		signal_data_pass1.push_back(signal_data_pass0[0]);
		init = true;
		
		for(iter = signal_data_pass0.begin() + 1;iter != signal_data_pass0.end() - 1;++iter)
		{
			curve_data[0] = (*(iter - 1));
			curve_data[1] = (*iter);
			curve_data[2] = (*(iter + 1));
			
			this->Get_Sub_Points(curve_data[0],curve_data[1],curve_data[2],&first_point,&second_point);
			
			if(init)
			{
				init = false;
			}
			else
			{
				first_point += last_second_point;
				first_point /= 2.0;
			}
			
			signal_data_pass1.push_back(first_point);
			signal_data_pass1.push_back(curve_data[1]);
			
			last_second_point = second_point;
		}

		if(!init)
		{
			signal_data_pass1.push_back(second_point);
			signal_data_pass1.push_back(curve_data[2]);
		}
		
		// second iteration
		signal_data_pass2.push_back(signal_data_pass1[0]);
		init = true;
		
		for(iter = signal_data_pass1.begin() + 1;iter != signal_data_pass1.end() - 1;++iter)
		{
			curve_data[0] = (*(iter - 1));
			curve_data[1] = (*iter);
			curve_data[2] = (*(iter + 1));
			
			this->Get_Sub_Points(curve_data[0],curve_data[1],curve_data[2],&first_point,&second_point);
			
			if(init)
			{
				init = false;
			}
			else
			{
				first_point += last_second_point;
				first_point /= 2.0;
			}
			
			signal_data_pass2.push_back(first_point);
			signal_data_pass2.push_back(curve_data[1]);
			
			last_second_point = second_point;
		}

		if(!init)
		{
			signal_data_pass2.push_back(second_point);
			signal_data_pass2.push_back(curve_data[2]);
		}
		
		angle_increment = PI2 / static_cast<double>(signal_data.size());
		
		for(iter = signal_data.begin(),angle = 0;iter != signal_data.end();++iter,angle += angle_increment)
		{
			if(Extrapolation::Get_Value(signal_data_pass2,angle,&dval))
			{
				pnt.Set(angle,dval);
			}
			else
			{
				pnt = (*iter);
			}
			
			normalized_signal.push_back(pnt);
		}
	}

	return normalized_signal;
}

void TFilterEvaluation::Get_Sub_Points(
	const TVector2 						&p1,
	const TVector2 						&p2,
	const TVector2 						&p3,
	TVector2 							* const first_pt,
	TVector2 							* const second_pt) const
{
	TVector2							pt;
	TVector2							v1,v2,v3;
	TVector3							p1_3,p2_3,p3_3;
	double								length_v1;
	double								length_v2;
	double								length_sample;
	double								length_dot;
	static const double					MIN_LENGTH(0.000001);
	
	assert(first_pt);
	assert(second_pt);
	
	*first_pt = (p1 + p2) / 2.0;
	*second_pt = (p2 + p3) / 2.0;
	
	v1 = p2 - p1;
	v2 = p3 - p2;
	
	length_v1 = v1.Length();
	length_v2 = v2.Length();
	
	if(length_v1 > MIN_LENGTH && length_v2 > MIN_LENGTH)
	{
		v3 = p3 - p1;
		v3.y = 0;
		length_sample = v3.Length();

		p1_3.Set(v1.x,v1.y,0.0);
		p2_3.Set(v2.x,v2.y,0.0);
		
		p3_3 = p2_3 * p1_3;
		
		pt = v1;
		
		v1.y = pt.x;
		v1.x = -pt.y;
		
		pt = v2;
		
		v2.y = pt.x;
		v2.x = -pt.y;
		
		v1.Normal();
		v2.Normal();
		
		if(p3_3.z < 0.0)
		{
			v1 *= (-1.0);
			v2 *= (-1.0);
		}
		
		length_dot = (v1 ^ v2);
		
		if(length_dot < 1.0)
		{
			length_dot = sqrt(1 - length_dot) / 4.0;
			length_dot *= (length_sample / 1.0);
		}
		else
		{
			length_dot = 0.0;
		}
		
		*first_pt += (v1 * length_dot * length_v1);
		*second_pt += (v2 * length_dot * length_v2);
	}
}

void TFilterEvaluation::Convert_To_Undulations(
	std::vector<TVector2> 				* const data) const
{
	std::vector<TVector2>::iterator		iter;

	assert(data);

	for(iter = data->begin();iter != data->end();++iter)
	{
		(*iter).x = this->Convert_Frequency_To_Undulations(data->size(),(*iter).x);
	}
}

double TFilterEvaluation::Convert_Frequency_To_Undulations(
	const int 							data_size,
	const double 						&frequency) const
{
	double								result(0);

	if(data_size)
	{
		result = (frequency / (static_cast<double>(data_size) / PI2)) * static_cast<double>(data_size);
	}
	
	return result;
}

int TFilterEvaluation::Get_Minimum_Sample_Size(
	const int 							max_undulations) const
{
	int									base;
	int									undulations_size(max_undulations * 2);
	int									size;
	
	base = log2(undulations_size);
	
	if(pow(2,base) == undulations_size)
	{
		size = undulations_size * 4;
	}
	else
	{
		size = pow(2,base + 1) * 4;
	}
	
	if(size < static_cast<int>(MINIMUM_SAMPLE_SIZE))
	{
		size = MINIMUM_SAMPLE_SIZE;
	}
	
	return size;
}

int TFilterEvaluation::Get_Sample_Window_Size(
	const int 							data_size) const
{
	int									base;
	
	assert(data_size > 0);
	
	base = log2(data_size);
	
	if(pow(2,base) == data_size)
	{
		return data_size;
	}

	return static_cast<int>(pow(2,base));
}

void TFilterEvaluation::Write_DMIS_Circle(
	const QString 						&file_name) const
{
	QFile								file;
	std::vector<TVector3>				pnts;
	std::vector<TVector3>::iterator 	pnt_iter;
	std::vector<TVector2>::const_iterator signal_iter;
	TReferenceSystem					refsys;
	QString								text;
	TVector3							position;
	TVector3							orientation;
	TVector3							pt;
	TVector3							vec;
	double								diameter;
	double								radius;
	int									cntr;
		
	if(!d_signal_data.size())
	{
		return;
	}
	
	position = d_feature_widget->Position();
	orientation = d_feature_widget->Orientation();
	diameter = d_feature_widget->Feature_Circle_Diameter();
	
	radius = diameter / 2.0;
	
	refsys.Skew1(orientation,TReferenceSystem::ZPLUS);
	refsys.Set_Origin(position);

	pnts.resize(d_signal_data.size());

	pt.z = 0.0;
	for(pnt_iter = pnts.begin(),signal_iter = d_signal_data.begin();
		pnt_iter != pnts.end();
		++pnt_iter,++signal_iter)
	{
		pt.x = cos((*signal_iter).x) * (radius + (*signal_iter).y);
		pt.y = sin((*signal_iter).x) * (radius + (*signal_iter).y);
		
		(*pnt_iter) = refsys.ToWorld(pt);
	}

	file.setFileName(file_name);
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{		
		// header
		file.write("$$ filterevaluation DMIS file\n");
		file.write("DMISMN/'filterevaluation.dms' 04.0\n");
		file.write("FILNAM/'filterevaluation.dms'\n");
		file.write("$$\n");
		file.write("UNITS/MM,ANGDEC\n");
		file.write("D(MC_CSY)=DATSET/MCS\n");
		file.write("MODE/MAN\n");
		file.write("$$\n");
		file.write("$$ Increase the display precision to 5 decimal places minimum.\n");
		file.write("DECPL/ALL,5\n");
		file.write("$$\n");
		
		file.write("MODE/PROG,MAN\n");
		file.write("$$\n");

		cntr = 1;
		for(pnt_iter = pnts.begin();pnt_iter != pnts.begin() + PADDING_SIZE;++pnt_iter,++cntr)
		{
			vec = position - (*pnt_iter);
			vec.Normal();
			
			text = QString("F(CIR_PNT_%1_PAD)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
					.arg(cntr)
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			text = QString("MEAS/POINT,F(CIR_PNT_%1_PAD),1\n")
						.arg(cntr);
			file.write(text.toLatin1());

			text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			file.write("ENDMES\n");
			file.write("$$\n");
		}

		for(pnt_iter = pnts.begin() + PADDING_SIZE;pnt_iter != pnts.end() - PADDING_SIZE;++pnt_iter,++cntr)
		{
			vec = position - (*pnt_iter);
			vec.Normal();
			
			text = QString("F(CIR_PNT_%1)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
					.arg(cntr)
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			text = QString("MEAS/POINT,F(CIR_PNT_%1),1\n")
						.arg(cntr);
			file.write(text.toLatin1());

			text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			file.write("ENDMES\n");
			file.write("$$\n");
		}

		for(pnt_iter = pnts.end() - PADDING_SIZE;pnt_iter != pnts.end();++pnt_iter,++cntr)
		{
			vec = position - (*pnt_iter);
			vec.Normal();
			
			text = QString("F(CIR_PNT_%1_PAD)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
					.arg(cntr)
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			text = QString("MEAS/POINT,F(CIR_PNT_%1_PAD),1\n")
						.arg(cntr);
			file.write(text.toLatin1());

			text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			file.write("ENDMES\n");
			file.write("$$\n");
		}

		file.write("ENDFIL\n\n");	
		file.close();
	}
	else
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Cannot open file for writing."));
		d_msg_box->setDetailedText(file_name);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
	}
}

void TFilterEvaluation::Write_DMIS_Cylinder(
	const QString 						&file_name) const
{
	QFile								file;
	std::vector<TVector3>				pnts_pad_start;
	std::vector<TVector3>				pnts_pad_end;
	std::vector<TVector3>				pnts;
	std::vector<TVector3>::iterator 	pnt_iter;
	std::vector<TVector2>::const_iterator signal_iter;
	TReferenceSystem					refsys;
	QString								text;
	TVector3							position;
	TVector3							orientation;
	TVector3							pt;
	TVector3							vec;
	double								diameter;
	double								radius;
	double								length;
	double								length_increment;
	double								length_offset;
	int									level_cntr;
	int									cntr;
	int									levels;
	static const double					FEATURE_LENGTH_MINIMUM(0.1);
		
	if(!d_signal_data.size())
	{
		return;
	}
	
	position = d_feature_widget->Position();
	orientation = d_feature_widget->Orientation();
	diameter = d_feature_widget->Feature_Cylinder_Diameter();
	length = d_feature_widget->Feature_Cylinder_Length();
	levels = d_feature_widget->Feature_Cylinder_Levels();
	
	if(length < FEATURE_LENGTH_MINIMUM)
	{
		return;
	}
		
	if(levels < 2)
	{
		return;
	}
	
	length_increment = length;
	length_increment /= static_cast<double>((d_signal_data.size() * levels) - 1);
	radius = diameter / 2.0;
	
	refsys.Skew1(orientation,TReferenceSystem::ZPLUS);
	refsys.Set_Origin(position - (orientation * length / 2.0));
	
	pnts_pad_start.resize(PADDING_SIZE);
	pnts.resize(d_signal_data.size() - PADDING_SIZE * 2);
	pnts_pad_end.resize(PADDING_SIZE);
	
	pt.z = 0.0;
	for(pnt_iter = pnts_pad_start.begin(),signal_iter = d_signal_data.begin();
		pnt_iter != pnts_pad_start.end();
		++pnt_iter,++signal_iter)
	{
		pt.x = cos((*signal_iter).x) * (radius + (*signal_iter).y);
		pt.y = sin((*signal_iter).x) * (radius + (*signal_iter).y);
		
		(*pnt_iter) = refsys.ToWorld(pt);
	}

	pt.z = 0.0;
	for(pnt_iter = pnts.begin(),signal_iter = d_signal_data.begin() + PADDING_SIZE;
		pnt_iter != pnts.end();
		++pnt_iter,++signal_iter)
	{
		pt.x = cos((*signal_iter).x) * (radius + (*signal_iter).y);
		pt.y = sin((*signal_iter).x) * (radius + (*signal_iter).y);
		
		(*pnt_iter) = refsys.ToWorld(pt);
	}
	
	pt.z = 0.0;
	for(pnt_iter = pnts_pad_end.begin(),signal_iter = d_signal_data.end() - PADDING_SIZE;
		pnt_iter != pnts_pad_end.end();
		++pnt_iter,++signal_iter)
	{
		pt.x = cos((*signal_iter).x) * (radius + (*signal_iter).y);
		pt.y = sin((*signal_iter).x) * (radius + (*signal_iter).y);
		
		(*pnt_iter) = refsys.ToWorld(pt);
	}

	file.setFileName(file_name);
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		// header
		file.write("$$ filterevaluation DMIS file\n");
		file.write("DMISMN/'filterevaluation.dms' 04.0\n");
		file.write("FILNAM/'filterevaluation.dms'\n");
		file.write("$$\n");
		file.write("UNITS/MM,ANGDEC\n");
		file.write("D(MC_CSY)=DATSET/MCS\n");
		file.write("MODE/MAN\n");
		file.write("$$\n");
		file.write("$$ Increase the display precision to 5 decimal places minimum.\n");
		file.write("DECPL/ALL,5\n");
		file.write("$$\n");
		
		file.write("MODE/PROG,MAN\n");
		file.write("$$\n");

		length_offset = 0;
		cntr = 1;
		for(level_cntr = 0;level_cntr < levels;++level_cntr)
		{
			if(level_cntr == 0)
			{
				for(pnt_iter = pnts_pad_start.begin();pnt_iter != pnts_pad_start.end();++pnt_iter,++cntr)
				{
					pt = (*pnt_iter);
					pt += (orientation * length_offset);
					
					vec = position;
					vec -= (orientation * length / 2.0);
					vec += (orientation * length_offset);
					vec -= pt;
					vec.Normal();
									
					text = QString("F(CYL_PNT_%1_PAD)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
							.arg(cntr)
							.arg(pt.x,0,'f',5)
							.arg(pt.y,0,'f',5)
							.arg(pt.z,0,'f',5)
							.arg(vec.x,0,'f',9)
							.arg(vec.y,0,'f',9)
							.arg(vec.z,0,'f',9);
					file.write(text.toLatin1());

					text = QString("MEAS/POINT,F(CYL_PNT_%1_PAD),1\n")
								.arg(cntr);
					file.write(text.toLatin1());

					text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
							.arg(pt.x,0,'f',5)
							.arg(pt.y,0,'f',5)
							.arg(pt.z,0,'f',5)
							.arg(vec.x,0,'f',9)
							.arg(vec.y,0,'f',9)
							.arg(vec.z,0,'f',9);
					file.write(text.toLatin1());

					file.write("ENDMES\n");
					file.write("$$\n");

					length_offset += length_increment;
				}
			}
			
			for(pnt_iter = pnts.begin();pnt_iter != pnts.end();++pnt_iter,++cntr)
			{
				pt = (*pnt_iter);
				pt += (orientation * length_offset);
				
				vec = position;
				vec -= (orientation * length / 2.0);
				vec += (orientation * length_offset);
				vec -= pt;
				vec.Normal();
								
				text = QString("F(CYL_PNT_%1)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
						.arg(cntr)
						.arg(pt.x,0,'f',5)
						.arg(pt.y,0,'f',5)
						.arg(pt.z,0,'f',5)
						.arg(vec.x,0,'f',9)
						.arg(vec.y,0,'f',9)
						.arg(vec.z,0,'f',9);
				file.write(text.toLatin1());

				text = QString("MEAS/POINT,F(CYL_PNT_%1),1\n")
							.arg(cntr);
				file.write(text.toLatin1());

				text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
						.arg(pt.x,0,'f',5)
						.arg(pt.y,0,'f',5)
						.arg(pt.z,0,'f',5)
						.arg(vec.x,0,'f',9)
						.arg(vec.y,0,'f',9)
						.arg(vec.z,0,'f',9);
				file.write(text.toLatin1());

				file.write("ENDMES\n");
				file.write("$$\n");

				length_offset += length_increment;
			}
		}
		
		for(pnt_iter = pnts_pad_end.begin();pnt_iter != pnts_pad_end.end();++pnt_iter,++cntr)
		{
			pt = (*pnt_iter);
			pt += (orientation * length_offset);
			
			vec = position;
			vec -= (orientation * length / 2.0);
			vec += (orientation * length_offset);
			vec -= pt;
			vec.Normal();
							
			text = QString("F(CYL_PNT_%1_PAD)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
					.arg(cntr)
					.arg(pt.x,0,'f',5)
					.arg(pt.y,0,'f',5)
					.arg(pt.z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			text = QString("MEAS/POINT,F(CYL_PNT_%1_PAD),1\n")
						.arg(cntr);
			file.write(text.toLatin1());

			text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
					.arg(pt.x,0,'f',5)
					.arg(pt.y,0,'f',5)
					.arg(pt.z,0,'f',5)
					.arg(vec.x,0,'f',9)
					.arg(vec.y,0,'f',9)
					.arg(vec.z,0,'f',9);
			file.write(text.toLatin1());

			file.write("ENDMES\n");
			file.write("$$\n");

			length_offset += length_increment;
		}

		
		file.write("ENDFIL\n\n");
		file.close();
	}
	else
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Cannot open file for writing."));
		d_msg_box->setDetailedText(file_name);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
	}
}

void TFilterEvaluation::Write_DMIS_Line(
	const QString 						&file_name) const
{
	QFile								file;
	std::vector<TVector3>				pnts;
	std::vector<TVector3>::iterator 	pnt_iter;
	std::vector<TVector2>::const_iterator signal_iter;
	TReferenceSystem					refsys;
	QString								text;
	TVector3							position;
	TVector3							orientation;
	TVector3							touch_vector;
	TVector3							pt;
	TVector3							vec;
	double								length;
	double								line_increment;
	double								line_pos;
	int									cntr;
	static const double					FEATURE_LENGTH_MINIMUM(10.0);
	
	if(!d_signal_data.size())
	{
		return;
	}
	
	position = d_feature_widget->Position();
	orientation = d_feature_widget->Orientation();
	length = d_feature_widget->Feature_Line_Length();
	
	if(length < FEATURE_LENGTH_MINIMUM)
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Line length too short."));
		d_msg_box->setDetailedText(file_name);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();

		return;
	}

	line_increment = length / static_cast<double>(d_signal_data.size() - PADDING_SIZE * 2);

	refsys.Skew1(orientation,TReferenceSystem::XPLUS);
	refsys.Set_Origin(position - (orientation * length / 2.0));

	pnts.resize(d_signal_data.size());
	
	line_pos = 0.0 - (static_cast<double>(PADDING_SIZE) * line_increment);

	for(pnt_iter = pnts.begin(),signal_iter = d_signal_data.begin();
		pnt_iter != pnts.end();
		++pnt_iter,++signal_iter,line_pos += line_increment)
	{
		pt.Set(line_pos,(*signal_iter).y,0.0);
		(*pnt_iter) = refsys.ToWorld(pt);
	}
	
	touch_vector = refsys.AxisToWorld(TVector3(0,1,0));

	file.setFileName(file_name);
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		// header
		file.write("$$ filterevaluation DMIS file\n");
		file.write("DMISMN/'filterevaluation.dms' 04.0\n");
		file.write("FILNAM/'filterevaluation.dms'\n");
		file.write("$$\n");
		file.write("UNITS/MM,ANGDEC\n");
		file.write("D(MC_CSY)=DATSET/MCS\n");
		file.write("MODE/MAN\n");
		file.write("$$\n");
		file.write("$$ Increase the display precision to 5 decimal places minimum.\n");
		file.write("DECPL/ALL,5\n");
		file.write("$$\n");
		
		file.write("MODE/PROG,MAN\n");
		file.write("$$\n");

		cntr = 1;
		for(pnt_iter = pnts.begin();pnt_iter != pnts.begin() + PADDING_SIZE;++pnt_iter,++cntr)
		{			
			text = QString("F(LN_%1_PAD)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
					.arg(cntr)
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(touch_vector.x,0,'f',9)
					.arg(touch_vector.y,0,'f',9)
					.arg(touch_vector.z,0,'f',9);
			file.write(text.toLatin1());

			text = QString("MEAS/POINT,F(LN_%1_PAD),1\n")
						.arg(cntr);
			file.write(text.toLatin1());

			text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(touch_vector.x,0,'f',9)
					.arg(touch_vector.y,0,'f',9)
					.arg(touch_vector.z,0,'f',9);
			file.write(text.toLatin1());

			file.write("ENDMES\n");
			file.write("$$\n");

		}
		for(pnt_iter = pnts.begin() + PADDING_SIZE;pnt_iter != pnts.end() - PADDING_SIZE;++pnt_iter,++cntr)
		{			
			text = QString("F(LN_%1)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
					.arg(cntr)
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(touch_vector.x,0,'f',9)
					.arg(touch_vector.y,0,'f',9)
					.arg(touch_vector.z,0,'f',9);
			file.write(text.toLatin1());

			text = QString("MEAS/POINT,F(LN_%1),1\n")
						.arg(cntr);
			file.write(text.toLatin1());

			text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(touch_vector.x,0,'f',9)
					.arg(touch_vector.y,0,'f',9)
					.arg(touch_vector.z,0,'f',9);
			file.write(text.toLatin1());

			file.write("ENDMES\n");
			file.write("$$\n");

		}
		for(pnt_iter = pnts.end() - PADDING_SIZE;pnt_iter != pnts.end();++pnt_iter,++cntr)
		{			
			text = QString("F(LN_%1_PAD)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7\n")
					.arg(cntr)
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(touch_vector.x,0,'f',9)
					.arg(touch_vector.y,0,'f',9)
					.arg(touch_vector.z,0,'f',9);
			file.write(text.toLatin1());

			text = QString("MEAS/POINT,F(LN_%1_PAD),1\n")
						.arg(cntr);
			file.write(text.toLatin1());

			text = QString("    PTMEAS/CART,%1,%2,%3,%4,%5,%6\n")
					.arg((*pnt_iter).x,0,'f',5)
					.arg((*pnt_iter).y,0,'f',5)
					.arg((*pnt_iter).z,0,'f',5)
					.arg(touch_vector.x,0,'f',9)
					.arg(touch_vector.y,0,'f',9)
					.arg(touch_vector.z,0,'f',9);
			file.write(text.toLatin1());

			file.write("ENDMES\n");
			file.write("$$\n");

		}
		
		file.write("ENDFIL\n\n");		
		file.close();
	}
	else
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Cannot open file for writing."));
		d_msg_box->setDetailedText(file_name);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
	}
}

void TFilterEvaluation::Write_Signal_Circle(
	const QString 						&file_name) const
{
	QFile								file;
	std::vector<TVector3>				pnts;
	std::vector<TVector3>::iterator 	pnt_iter;
	std::vector<TVector2>::const_iterator signal_iter;
	TReferenceSystem					refsys;
	QString								text;
	TVector3							position;
	TVector3							orientation;
	TVector3							pt;
	double								diameter;
	double								radius;
	int									cntr;
		
	if(!d_signal_data.size())
	{
		return;
	}
	
	position = d_feature_widget->Position();
	orientation = d_feature_widget->Orientation();
	diameter = d_feature_widget->Feature_Circle_Diameter();
	
	radius = diameter / 2.0;
	
	refsys.Skew1(orientation,TReferenceSystem::ZPLUS);
	refsys.Set_Origin(position);
	
	pnts.resize(d_signal_data.size());

	pt.z = 0.0;
	for(pnt_iter = pnts.begin(),signal_iter = d_signal_data.begin();
		pnt_iter != pnts.end();
		++pnt_iter,++signal_iter)
	{
		pt.x = cos((*signal_iter).x) * (radius + (*signal_iter).y);
		pt.y = sin((*signal_iter).x) * (radius + (*signal_iter).y);
		
		(*pnt_iter) = refsys.ToWorld(pt);
	}

	file.setFileName(file_name);
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		// header
		file.write("Filter_Evaluation_Signal_Data:Version=1:Type=Signal_Data\n\n");
		file.write("# First line must start with 'Filter_Evaluation_Signal_Data'\n");
		file.write("# 'Version=<n>' is the revision of this file\n");
		file.write("# The value of Type must be 'Signal_Data'.  Other options may be added in the future.\n\n");
		file.write("# Data is grouped in pairs <desc>_Begin: and <desc>_End:\n\n");
		file.write("# Position and Orientation entries must exist.\n");
		file.write("# For circular features the diameter must also exist.\n\n");
		file.write("# Signal type can be either 'Circle' or 'Line'.\n\n");
		file.write("# Start and end of the data is extended or padded.\n\n");
		file.write("Signal_Type:Circle\n");
		
		text = QString("Position:%1,%2,%3\n")
				.arg(position.x,0,'f',5)
				.arg(position.y,0,'f',5)
				.arg(position.z,0,'f',5);
		file.write(text.toLatin1());
		
		text = QString("Orientation:%1,%2,%3\n")
				.arg(orientation.x,0,'f',5)
				.arg(orientation.y,0,'f',5)
				.arg(orientation.z,0,'f',5);
		file.write(text.toLatin1());
		
		text = QString("Diameter:%1\n\n").arg(diameter,0,'f',5);
		file.write(text.toLatin1());

		file.write("Data_Begin:\n");

		for(pnt_iter = pnts.begin(),cntr = 0;
			pnt_iter != pnts.end();
			++pnt_iter,++cntr)
		{
			text = QString("%1,%2,%3\n")
						.arg((*pnt_iter).x,0,'f',5)
						.arg((*pnt_iter).y,0,'f',5)
						.arg((*pnt_iter).z,0,'f',5);
			file.write(text.toLatin1());
		}
		
		file.write("Data_End:\n");

		
		file.close();
	}
	else
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Cannot open file for writing."));
		d_msg_box->setDetailedText(file_name);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
	}
}

void TFilterEvaluation::Write_Signal_Line(
	const QString 						&file_name) const
{
	QFile								file;
	std::vector<TVector3>				pnts;
	std::vector<TVector3>::iterator 	pnt_iter;
	std::vector<TVector2>::const_iterator signal_iter;
	TReferenceSystem					refsys;
	QString								text;
	TVector3							position;
	TVector3							orientation;
	TVector3							pt;
	double								length;
	double								line_pos;
	double								line_increment;
	static const double					FEATURE_LENGTH_MINIMUM(10.0);

	if(!d_signal_data.size())
	{
		return;
	}
	
	position = d_feature_widget->Position();
	orientation = d_feature_widget->Orientation();
	length = d_feature_widget->Feature_Line_Length();
	
	if(length < FEATURE_LENGTH_MINIMUM)
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Line length too short."));
		d_msg_box->setDetailedText(file_name);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();

		return;
	}

	line_increment = length / static_cast<double>(d_signal_data.size() - PADDING_SIZE * 2);

	refsys.Skew1(orientation,TReferenceSystem::XPLUS);
	refsys.Set_Origin(position - (orientation * length / 2.0));

	pnts.resize(d_signal_data.size());

	line_pos = 0.0 - (static_cast<double>(PADDING_SIZE) * line_increment);

	for(pnt_iter = pnts.begin(),signal_iter = d_signal_data.begin();
		pnt_iter != pnts.end();
		++pnt_iter,++signal_iter,line_pos += line_increment)
	{
		pt.Set(line_pos,(*signal_iter).y,0.0);
		(*pnt_iter) = refsys.ToWorld(pt);
	}

	file.setFileName(file_name);
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		// header
		file.write("Filter_Evaluation_Signal_Data:Version=1:Type=Signal_Data\n\n");
		file.write("# First line must start with 'Filter_Evaluation_Signal_Data'\n");
		file.write("# 'Version=<n>' is the revision of this file\n");
		file.write("# The value of Type must be 'Signal_Data'.  Other options may be added in the future.\n\n");
		file.write("# Data is grouped in pairs <desc>_Begin: and <desc>_End:\n\n");
		file.write("# Position and Orientation entries must exist.\n");
		file.write("# For circular features the diameter must also exist.\n\n");
		file.write("# Signal type can be either 'Circle' or 'Line'.\n\n");
		file.write("# Start and end of the data is extended or padded.\n\n");
		file.write("Signal_Type:Line\n");
		
		text = QString("Position:%1,%2,%3\n")
				.arg(position.x,0,'f',5)
				.arg(position.y,0,'f',5)
				.arg(position.z,0,'f',5);
		file.write(text.toLatin1());
		
		text = QString("Orientation:%1,%2,%3\n")
				.arg(orientation.x,0,'f',5)
				.arg(orientation.y,0,'f',5)
				.arg(orientation.z,0,'f',5);
		file.write(text.toLatin1());
		
		pt = refsys.AxisToWorld(TVector3(0,1,0));

		text = QString("Signal_Axis:%1,%2,%3\n\n")
				.arg(pt.x,0,'f',5)
				.arg(pt.y,0,'f',5)
				.arg(pt.z,0,'f',5);
		file.write(text.toLatin1());

		file.write("Data_Begin:\n");

		for(pnt_iter = pnts.begin();pnt_iter != pnts.end();++pnt_iter)
		{
			text = QString("%1,%2,%3\n")
						.arg((*pnt_iter).x,0,'f',5)
						.arg((*pnt_iter).y,0,'f',5)
						.arg((*pnt_iter).z,0,'f',5);
			file.write(text.toLatin1());
		}
		
		file.write("Data_End:\n");

		
		file.close();
	}
	else
	{
		d_msg_box->setText(QStringLiteral("Error"));
		d_msg_box->setInformativeText(QStringLiteral("Cannot open file for writing."));
		d_msg_box->setDetailedText(file_name);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
	}
}

std::vector<TVector2> TFilterEvaluation::Read_Signal_File(
	const QString 						&file_name,
	bool 								* const valid,
	TFilterEvaluation::TSignalType 		* const signal_type,
	double 								* const range) const
{
	std::vector<TVector2>				signal_data;
	std::vector<TVector2>::iterator		iter;
	TReferenceSystem					refsys;
	QByteArray							file_data;
	QFile								file;
	QString								line;
	QString								text;
	QStringList							line_items;
	TVector2							signal_pnt;
	TVector2							signal_pnt_min;
	TVector2							signal_pnt_max;
	TVector3							pnt;
	TVector3							pc;
	TVector3							vec;
	TVector3							position;
	TVector3							orientation;
	TVector3							signal_axis;
	double								diameter(0.0);
	double								pos_multiplier;
	bool								found_position(false);
	bool								found_orientation(false);
	bool								found_signal_axis(false);
	bool								found_diameter(false);
	bool								found_type(false);
	int									sample_window_size;
	bool								inside_data(false);
	bool								init(true);

	assert(valid);
	assert(signal_type);
	assert(range);
	
	*valid = false;
	*signal_type = TFilterEvaluation::UNKNOWN_TYPE;
	*range = 0.0;
	
	orientation.Set(0,0,1);
	signal_axis.Set(0,1,0);

	file.setFileName(file_name);
	if(file.open(QIODevice::ReadOnly | QIODevice::Text))
	{
		file_data = file.readAll();
		file.close();
	}
	
	if(file_data.size())
	{
		QTextStream						stream(file_data);

		while(!stream.atEnd())
		{
			line = stream.readLine().simplified();
						
			if(line.size() > 0 && line.startsWith('#') == false)
			{
				if(init)
				{
					if(line.compare(QStringLiteral("Filter_Evaluation_Signal_Data:Version=1:Type=Signal_Data")) != 0)
					{
						d_last_error = QStringLiteral("Invalid header line");
						return signal_data;
					}
					
					init = false;
				}
				else
				{
					if(!inside_data)
					{
						if(line.startsWith(QStringLiteral("Signal_Type:"),Qt::CaseInsensitive))
						{
							text = line.mid(12);	// remove Signal_Type:
							
							if(text.compare(QStringLiteral("Circle"),Qt::CaseInsensitive) == 0)
							{
								*signal_type = TFilterEvaluation::CIRCLE_SIGNAL;
							}
							else if(text.compare(QStringLiteral("Line"),Qt::CaseInsensitive) == 0)
							{
								*signal_type = TFilterEvaluation::LINE_SIGNAL;
							}

							found_type = true;
						}
						else if(line.startsWith(QStringLiteral("Position:"),Qt::CaseInsensitive))
						{
							text = line.mid(9);	// remove Position:
							
							line_items = text.split(',');
							
							if(line_items.size() == 3)
							{
								position.x = line_items[0].toDouble();
								position.y = line_items[1].toDouble();
								position.z = line_items[2].toDouble();

								found_position = true;
							}
						}
						else if(line.startsWith(QStringLiteral("Orientation:"),Qt::CaseInsensitive))
						{
							text = line.mid(12);	// remove Orientation:
							
							line_items = text.split(',');
							
							if(line_items.size() == 3)
							{
								orientation.x = line_items[0].toDouble();
								orientation.y = line_items[1].toDouble();
								orientation.z = line_items[2].toDouble();
								
								if(orientation.Length() > ZERO_EPSILON)
								{
									orientation.Normal();
								}
								else
								{
									orientation.Set(0,0,1);
								}

								found_orientation = true;
							}
						}
						else if(line.startsWith(QStringLiteral("Signal_Axis:"),Qt::CaseInsensitive))
						{
							text = line.mid(12);	// remove Signal_Axis:
							
							line_items = text.split(',');
							
							if(line_items.size() == 3)
							{
								signal_axis.x = line_items[0].toDouble();
								signal_axis.y = line_items[1].toDouble();
								signal_axis.z = line_items[2].toDouble();

								found_signal_axis = true;
							}
						}
						else if(line.startsWith(QStringLiteral("Diameter:"),Qt::CaseInsensitive))
						{
							text = line.mid(9);	// remove Diameter:
							
							diameter = text.toDouble();

							found_diameter = true;
						}
						else if(line.startsWith(QStringLiteral("Data_Begin:"),Qt::CaseInsensitive))
						{
							refsys.Skew1(orientation,TReferenceSystem::ZPLUS);
							refsys.Set_Origin(position);

							inside_data = true;
						}
						else if(line.startsWith(QStringLiteral("Data_End:"),Qt::CaseInsensitive))
						{
							inside_data = false;
						}
					}
					else
					{
						line_items = line.split(',');
						
						if(line_items.size() == 3)
						{
							pnt.x = line_items[0].toDouble();
							pnt.y = line_items[1].toDouble();
							pnt.z = line_items[2].toDouble();
														
							pnt = refsys.FromWorld(pnt);
							
							if((*signal_type) == TFilterEvaluation::CIRCLE_SIGNAL)
							{
								pnt.z = 0.0;
								vec = pnt;
								
								if(vec.Length() > ZERO_EPSILON)
								{
									vec.Normal();
																										
									signal_pnt.y = pnt.Length() - diameter / 2.0;
									
									if(vec.y < 0.0)
									{
										signal_pnt.x = PI2 - acos(vec.x);
									}
									else
									{
										signal_pnt.x = acos(vec.x);
									}
								}
							}
							else if((*signal_type) == TFilterEvaluation::LINE_SIGNAL)
							{
								vec = refsys.AxisFromWorld(signal_axis);
								pc = refsys.FromWorld(position);

								signal_pnt.x = pnt.z;
								
								pnt.z = 0;
								
								signal_pnt.y = (pnt.x - pc.x) * vec.x + (pnt.y - pc.y) * vec.y + (pnt.z - pc.z) * vec.z;
							}
																
							signal_data.push_back(signal_pnt);
						}
					}
				}
			}
		}
	}
		
	if(found_position && found_orientation && found_type)
	{
		if((*signal_type) == TFilterEvaluation::CIRCLE_SIGNAL)
		{
			*valid = found_diameter;
		}
		else
		{
			*valid = found_signal_axis;
		}
	}
		
	if(signal_data.size() < 2)
	{
		d_last_error = QStringLiteral("Size of signal data too small");
		*valid = false;
	}
	
	if(!(*valid))
	{
		d_last_error = QString("Found_Position: %1  Found_Orientation: %2  Found_Type: %3  Found_Diameter: %4  Found_Signal_Axis: %5")
					.arg(found_position)
					.arg(found_orientation)
					.arg(found_type)
					.arg(found_diameter)
					.arg(found_signal_axis);

	}
	else
	{
		
		sample_window_size = this->Get_Sample_Window_Size(static_cast<int>(signal_data.size()));

		if(signal_data.size() % 2)
		{
			signal_data.pop_back();
		}
		
		while(static_cast<int>(signal_data.size()) > sample_window_size)
		{
			signal_data.erase(signal_data.begin());
			signal_data.pop_back();
		};

		
		init = true;
		for(iter = signal_data.begin();iter != signal_data.end();++iter)
		{
			if(init)
			{
				signal_pnt_min = (*iter);
				signal_pnt_max = (*iter);

				init = false;
			}
			else
			{
				if((*iter).x < signal_pnt_min.x) {signal_pnt_min = (*iter);}
				if((*iter).x > signal_pnt_max.x) {signal_pnt_max = (*iter);}
			}
		}
		
		*range = signal_pnt_max.x - signal_pnt_min.x;
		*range += (*range) / static_cast<double>(signal_data.size() - 1);
		
		if((*signal_type) == TFilterEvaluation::LINE_SIGNAL && (*range) > ZERO_EPSILON && signal_data.size() > 1)
		{
			pos_multiplier = PI2 / (*range);
			
			for(iter = signal_data.begin();iter != signal_data.end();++iter)
			{
				(*iter).x -= signal_pnt_min.x;
				(*iter).x *= pos_multiplier;
			}
		}
	}

	return signal_data;
}

void TFilterEvaluation::Add_Signal_Entry(
	const int 							undulations,
	const double						&phase,
	const double 						&amplitude)
{
	std::vector<TVector2>::iterator		iter;
	double								dval;
	
	for(iter = d_signal_data.begin();iter != d_signal_data.end();++iter)
	{
		dval = sin((*iter).x * static_cast<double>(undulations) + phase) * amplitude;
		(*iter).y += dval;
	}
}
