/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QDoubleSpinBox>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QSpinBox>

#include "parameterdialog.h"

TParameterDialog::TParameterDialog(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*separator_hline;
	QGridLayout							*dialog_layout;
	QHBoxLayout							*button_hlayout;
	QLabel								*amplitude_label;
	QLabel								*phase_label;
	QLabel								*undulations_label;
	QSizePolicy							size_policy(QSizePolicy::MinimumExpanding,QSizePolicy::Fixed);
	QSpacerItem							*button_hspacer;

	size_policy.setHorizontalStretch(0);
	size_policy.setVerticalStretch(0);

	dialog_layout = new QGridLayout(this);

	undulations_label = new QLabel(this);
	dialog_layout->addWidget(undulations_label,0,0,1,1);

	d_undulations_spin = new QSpinBox(this);
	d_undulations_spin->setSizePolicy(size_policy);
	d_undulations_spin->setAlignment(Qt::AlignCenter);
	d_undulations_spin->setMinimum(1);
	d_undulations_spin->setMaximum(100);
	dialog_layout->addWidget(d_undulations_spin,0,1,1,1);

	phase_label = new QLabel(this);
	dialog_layout->addWidget(phase_label,1,0,1,1);

	d_phase_spin = new QDoubleSpinBox(this);
	d_phase_spin->setSizePolicy(size_policy);
	d_phase_spin->setAlignment(Qt::AlignCenter);
	d_phase_spin->setDecimals(2);
	d_phase_spin->setMinimum(-90.0);
	d_phase_spin->setMaximum(90.0);
	dialog_layout->addWidget(d_phase_spin,1,1,1,1);

	amplitude_label = new QLabel(this);
	dialog_layout->addWidget(amplitude_label,2,0,1,1);

	d_amplitude_spin = new QDoubleSpinBox(this);
	d_amplitude_spin->setSizePolicy(size_policy);
	d_amplitude_spin->setAlignment(Qt::AlignCenter);
	d_amplitude_spin->setDecimals(2);
	d_amplitude_spin->setMinimum(0.0);
	d_amplitude_spin->setMaximum(1.0);
	dialog_layout->addWidget(d_amplitude_spin,2,1,1,1);

	separator_hline = new QFrame(this);
	separator_hline->setFrameShape(QFrame::HLine);
	separator_hline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_hline,3,0,1,2);

	button_hlayout = new QHBoxLayout();

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	d_cancel_button = new QPushButton(this);
	d_cancel_button->setAutoDefault(false);
	button_hlayout->addWidget(d_cancel_button);

	d_accept_button = new QPushButton(this);
	d_accept_button->setAutoDefault(true);
	button_hlayout->addWidget(d_accept_button);
	dialog_layout->addLayout(button_hlayout,4,0,1,2);
	
	// defaults
	d_amplitude_spin->setValue(0.01);
	d_phase_spin->setValue(0.0);
	d_undulations_spin->setValue(10);
	d_current_mode = TParameterDialog::MODE_EDIT;

	undulations_label->setText(QStringLiteral("Undulations:"));
	phase_label->setText(QStringLiteral("Phase:"));
	amplitude_label->setText(QStringLiteral("Amplitude:"));

	connect(d_cancel_button,&QPushButton::clicked,this,&TParameterDialog::reject);
	connect(d_accept_button,&QPushButton::clicked,this,&TParameterDialog::Accept);
}

TParameterDialog::~TParameterDialog(void)
{
}

double TParameterDialog::Amplitude(void) const
{
	return d_amplitude_spin->value();
}

double TParameterDialog::Phase(void) const
{
	return d_phase_spin->value();
}

int TParameterDialog::Undulations(void) const
{
	return d_undulations_spin->value();
}

void TParameterDialog::Reset(
	const TParameterDialog::TMode 		mode)
{
	QString								text;
	
	d_current_mode = mode;
	
	switch(d_current_mode)
	{
		case TParameterDialog::MODE_ADD:
			text = QStringLiteral("Add Parameter Entry");

			d_cancel_button->setText(QStringLiteral("Close"));
			d_accept_button->setText(QStringLiteral("Add"));
			
			d_undulations_spin->setEnabled(true);

			break;
			
		case TParameterDialog::MODE_EDIT:
			text = QStringLiteral("Edit Parameter Entry");
			
			d_cancel_button->setText(QStringLiteral("Cancel"));
			d_accept_button->setText(QStringLiteral("Accept"));
			
			d_undulations_spin->setEnabled(true);

			break;
			
		case TParameterDialog::MODE_MULTI_EDIT:
			text = QStringLiteral("Edit Parameters Amplitude");
			
			d_cancel_button->setText(QStringLiteral("Cancel"));
			d_accept_button->setText(QStringLiteral("Accept"));
			
			d_undulations_spin->setEnabled(false);

			break;
	}
	
	this->setWindowTitle(text);
}

void TParameterDialog::Set_Amplitude(
	const double 						&value)
{
	d_amplitude_spin->setValue(value);
}

void TParameterDialog::Set_Phase(
	const double 						&value)
{
	d_phase_spin->setValue(value);
}

void TParameterDialog::Set_Undulations(
	const int 							value)
{
	d_undulations_spin->setValue(value);
}

void TParameterDialog::Accept(void)
{
	int									undulations;
	double								amplitude;
	double								phase;
	
	switch(d_current_mode)
	{
		case TParameterDialog::MODE_ADD:
			
			undulations = d_undulations_spin->value();
			amplitude = d_amplitude_spin->value();
			phase = d_phase_spin->value();

			emit Add_Signal_Entry(undulations,phase,amplitude);
			
			break;
			
		case TParameterDialog::MODE_EDIT:
		case TParameterDialog::MODE_MULTI_EDIT:
			this->accept();
			break;
	}
}


