/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QDoubleSpinBox>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QListWidget>
#include <QListWidgetItem>
#include <QSpinBox>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QToolButton>
#include <QIcon>
#include <QVariant>
#include <QList>
#include <QFile>
#include <map>
#include <algorithm>

#include "../../core/xmlfile.h"

#include "parameterdialog.h"

#include "parametersettingswidget.h"

Q_DECLARE_METATYPE(TParameterSettingsWidget::TEntryItem);

static bool EntryItem_Sort_Function(
	TParameterSettingsWidget::TEntryItem a,
	TParameterSettingsWidget::TEntryItem b)
{	
	return a < b;
}

TParameterSettingsWidget::TParameterSettingsWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QHBoxLayout							*button_hlayout;
	QLabel								*modulation_label;
	QLabel								*modulation_units_label;
	QLabel								*entries_label;
	QSizePolicy							sizepolicy_minexpanding_fixed(QSizePolicy::MinimumExpanding,QSizePolicy::Fixed);
	QSpacerItem							*button_hspacer;

	sizepolicy_minexpanding_fixed.setHorizontalStretch(0);
	sizepolicy_minexpanding_fixed.setVerticalStretch(0);

	widget_layout = new QGridLayout(this);
	widget_layout->setSpacing(4);
	widget_layout->setContentsMargins(4,4,4,4);
	
	modulation_label = new QLabel(this);
	widget_layout->addWidget(modulation_label,0,0,1,1);

	d_modulation_spin = new QSpinBox(this);
	d_modulation_spin->setAlignment(Qt::AlignCenter);
	d_modulation_spin->setMinimum(0);
	d_modulation_spin->setMaximum(100);
	d_modulation_spin->setSizePolicy(QSizePolicy::MinimumExpanding,QSizePolicy::Preferred);
	widget_layout->addWidget(d_modulation_spin,0,1,1,1);
	
	modulation_units_label = new QLabel(this);
	widget_layout->addWidget(modulation_units_label,0,2,1,1);

	entries_label = new QLabel(this);
	entries_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	widget_layout->addWidget(entries_label,1,0,1,1);

	d_entries_list = new QListWidget(this);
	d_entries_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	widget_layout->addWidget(d_entries_list,1,1,1,2);

	button_hlayout = new QHBoxLayout();

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	d_add_button = new QToolButton(this);
	d_add_button->setFixedSize(16,16);
	button_hlayout->addWidget(d_add_button);

	d_remove_button = new QToolButton(this);
	d_remove_button->setFixedSize(16,16);
	button_hlayout->addWidget(d_remove_button);

	d_edit_button = new QToolButton(this);
	d_edit_button->setFixedSize(16,16);
	button_hlayout->addWidget(d_edit_button);
	widget_layout->addLayout(button_hlayout,2,1,1,2);

	// defaults
	d_parameter_dialog = 0;
	d_add_button->setEnabled(true);
	d_remove_button->setEnabled(false);
	d_edit_button->setEnabled(false);
	d_modulation_spin->setValue(0);
	d_data_changed = false;

	entries_label->setText(QStringLiteral("Entries:"));
	modulation_label->setText(QStringLiteral("Modulation:"));
	modulation_units_label->setText(QStringLiteral("%"));

	d_add_button->setIcon(QIcon(":/button/add_x16.png"));
	d_remove_button->setIcon(QIcon(":/button/remove_x16.png"));
	d_edit_button->setIcon(QIcon(":/button/edit_x16.png"));

	connect(d_add_button,&QToolButton::clicked,this,&TParameterSettingsWidget::Add_Entry);
	connect(d_remove_button,&QToolButton::clicked,this,&TParameterSettingsWidget::Remove_Entries);
	connect(d_edit_button,&QToolButton::clicked,this,&TParameterSettingsWidget::Edit_Entry);
	connect(d_entries_list,&QListWidget::itemSelectionChanged,this,&TParameterSettingsWidget::Entry_Selection_Changed);
	connect(d_entries_list,&QListWidget::itemDoubleClicked,this,&TParameterSettingsWidget::Edit_Entry);
	connect(d_modulation_spin,static_cast<void (QSpinBox::*)(int)>(&QSpinBox::valueChanged),this,&TParameterSettingsWidget::Parameters_Changed);
}

TParameterSettingsWidget::~TParameterSettingsWidget(void)
{
}

bool TParameterSettingsWidget::Save_Data(
	const QString						&data_path) const
{
	if(d_data_changed)
	{
		if(this->Save(data_path))
		{
			d_data_changed = false;
			return true;
		}
		
		return false;
	}
	
	return true;
}

int TParameterSettingsWidget::Maximum_Undulations(void) const
{
	int									max(0);
	std::vector<TParameterSettingsWidget::TEntryItem>::const_iterator iter;
	
	for(iter = d_entry_items.begin();iter != d_entry_items.end();++iter)
	{
		if((*iter).undulations > max)
		{
			max = (*iter).undulations;
		}
	}

	return max;
}

double TParameterSettingsWidget::Maximum_Amplitude(void) const
{
	double								max(0.0);
	std::map<int,double>				map;
	std::map<int,double>::iterator		map_iter;
	std::vector<TParameterSettingsWidget::TEntryItem>::const_iterator iter;
	
	for(iter = d_entry_items.begin();iter != d_entry_items.end();++iter)
	{		
		map_iter = map.find((*iter).undulations);
		
		if(map_iter == map.end())
		{
			map[(*iter).undulations] = (*iter).amplitude;
		}
		else
		{
			(*map_iter).second += (*iter).amplitude;
		}
	}
	
	for(map_iter = map.begin();map_iter != map.end();++map_iter)
	{
		if((*map_iter).second > max)
		{
			max = (*map_iter).second;
		}
	}

	return max;
}

int TParameterSettingsWidget::Modulation(void) const
{
	return d_modulation_spin->value();
}

bool TParameterSettingsWidget::Load_Data(
	const QString						&data_path)
{
	TXmlFile							xmlfile;
	QString								file_name;
	int									version;
	
	file_name = QString("%1/parameters.xml").arg(data_path);
	
	if(!QFile::exists(file_name))
	{
		this->Generate_Test_Data();
		this->Update_List_From_Entry_Items();
		
		d_data_changed = true;
		return true;
	}
	
	try
	{
		xmlfile.Set_File_Name(file_name);
		xmlfile.Open("Parameters",&version);
		
		switch(version)
		{
			case 1:
				this->Load_Version1(&xmlfile);
				break;
		}
		
		this->Update_List_From_Entry_Items();
		d_data_changed = false;
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Loading machines exception: %1").arg(xml_exception.ErrorText());
		
		d_entries_list->clear();
		
		return false;
	}
	
	return true;
}

void TParameterSettingsWidget::Set_Modulation(
	const int 							ival)
{
	d_modulation_spin->setValue(ival);
}

void TParameterSettingsWidget::Add_Entry(void)
{	
	if(!d_parameter_dialog)
	{
		d_parameter_dialog = new TParameterDialog(this);
		connect(d_parameter_dialog,&TParameterDialog::Add_Signal_Entry,this,&TParameterSettingsWidget::Add_Signal_Entry);
	}
	
	d_parameter_dialog->Reset(TParameterDialog::MODE_ADD);
	
	d_parameter_dialog->exec();
}

void TParameterSettingsWidget::Remove_Entries(void)
{
	QList<QListWidgetItem*>				items;
	QList<QListWidgetItem*>::iterator	iter;
	int									row;
	
	d_remove_button->setEnabled(false);
	d_edit_button->setEnabled(false);
	
	items = d_entries_list->selectedItems();
	
	for(iter = items.begin();iter != items.end();++iter)
	{
		row = d_entries_list->row((*iter));
		d_entries_list->takeItem(row);
		
		delete (*iter);
	}
	
	this->Update_Entry_Items_From_List();
	
	d_data_changed = true;
	
	emit Parameters_Changed();
}

void TParameterSettingsWidget::Edit_Entry(void)
{
	QList<QListWidgetItem*>				items;
	QList<QListWidgetItem*>::iterator	iter;
	QListWidgetItem						*item;
	QVariant							variant;
	TParameterSettingsWidget::TEntryItem entry;
	double								amplitude;
	double								phase;
	
	items = d_entries_list->selectedItems();

	if(!d_parameter_dialog)
	{
		d_parameter_dialog = new TParameterDialog(this);
		connect(d_parameter_dialog,&TParameterDialog::Add_Signal_Entry,this,&TParameterSettingsWidget::Add_Signal_Entry);
	}
	
	if(items.size() == 1)
	{
		item = items[0];
		
		variant = item->data(Qt::UserRole);
		entry = variant.value<TParameterSettingsWidget::TEntryItem>();

		d_parameter_dialog->Reset(TParameterDialog::MODE_EDIT);
		
		d_parameter_dialog->Set_Amplitude(entry.amplitude);
		d_parameter_dialog->Set_Phase(entry.phase);
		d_parameter_dialog->Set_Undulations(entry.undulations);

		if(d_parameter_dialog->exec() == QDialog::Accepted)
		{
			entry.amplitude = d_parameter_dialog->Amplitude();
			entry.phase = d_parameter_dialog->Phase();
			entry.undulations = d_parameter_dialog->Undulations();
			
			variant.setValue(entry);
			item->setData(Qt::UserRole,variant);
			
			item->setText(QString("Undulations: %1 Ph: %2 Amp: %3").arg(entry.undulations).arg(entry.phase,0,'f',2).arg(entry.amplitude,0,'f',3));
					
			this->Update_Entry_Items_From_List();
			this->Update_List_From_Entry_Items();	// sort visible items
			
			d_data_changed = true;
			
			emit Parameters_Changed();
		}
	}
	else if (items.size() > 1)
	{
		item = items[0];	// use value from the first item
		
		variant = item->data(Qt::UserRole);
		entry = variant.value<TParameterSettingsWidget::TEntryItem>();

		d_parameter_dialog->Reset(TParameterDialog::MODE_MULTI_EDIT);
		
		d_parameter_dialog->Set_Amplitude(entry.amplitude);
		d_parameter_dialog->Set_Phase(entry.phase);
		d_parameter_dialog->Set_Undulations(entry.undulations);

		if(d_parameter_dialog->exec() == QDialog::Accepted)
		{
			amplitude = d_parameter_dialog->Amplitude();
			phase = d_parameter_dialog->Phase();

			for(iter = items.begin();iter != items.end();++iter)
			{
				variant = (*iter)->data(Qt::UserRole);
				entry = variant.value<TParameterSettingsWidget::TEntryItem>();

				entry.amplitude = amplitude;
				entry.phase = phase;
				
				variant.setValue(entry);
				(*iter)->setData(Qt::UserRole,variant);
				
				(*iter)->setText(QString("Undulations: %1 Ph: %2 Amp: %3").arg(entry.undulations).arg(entry.phase,0,'f',2).arg(entry.amplitude,0,'f',3));

			}
					
			this->Update_Entry_Items_From_List();
			this->Update_List_From_Entry_Items();	// sort visible items
			
			d_data_changed = true;
			
			emit Parameters_Changed();
		}
	}
}

void TParameterSettingsWidget::Entry_Selection_Changed(void)
{
	QList<QListWidgetItem*>				items;

	items = d_entries_list->selectedItems();
	
	d_remove_button->setEnabled(items.size() > 0);
	d_edit_button->setEnabled(items.size() > 0);
}

void TParameterSettingsWidget::Add_Signal_Entry(
	const int				 			undulations,
	const double						&phase,
	const double 						&amplitude)
{
	QVariant							variant;
	TParameterSettingsWidget::TEntryItem entry;
	
	entry.undulations = undulations;
	entry.amplitude = amplitude;
	entry.phase = phase;

	d_entry_items.push_back(entry);
	
	this->Update_List_From_Entry_Items();
	
	d_data_changed = true;
	
	emit Parameters_Changed();
}

void TParameterSettingsWidget::Generate_Test_Data(void)
{
	TParameterSettingsWidget::TEntryItem entry;
		
	d_entry_items.clear();

	entry.amplitude = 0.25;
	entry.phase = 0.0;
	entry.undulations = 10;
	d_entry_items.push_back(entry);

	entry.amplitude = 0.25;
	entry.phase = 0.0;
	entry.undulations = 20;
	d_entry_items.push_back(entry);

	entry.amplitude = 0.25;
	entry.phase = 0.0;
	entry.undulations = 30;
	d_entry_items.push_back(entry);

	entry.amplitude = 0.25;
	entry.phase = 0.0;
	entry.undulations = 40;
	d_entry_items.push_back(entry);

	entry.amplitude = 0.25;
	entry.phase = 0.0;
	entry.undulations = 50;
	d_entry_items.push_back(entry);
}

void TParameterSettingsWidget::Update_Entry_Items_From_List(void)
{
	TParameterSettingsWidget::TEntryItem entry;
	QListWidgetItem						*item;
	QVariant							variant;
	int									row;
	int									count;
	
	d_entry_items.clear();
	
	count = d_entries_list->count();
	
	for(row = 0;row < count;++row)
	{
		item = d_entries_list->item(row);
		
		variant = item->data(Qt::UserRole);
		entry = variant.value<TParameterSettingsWidget::TEntryItem>();
		
		d_entry_items.push_back(entry);
	}
	
	std::sort(d_entry_items.begin(), d_entry_items.end(), EntryItem_Sort_Function);
}

void TParameterSettingsWidget::Update_List_From_Entry_Items(void)
{
	std::vector<TParameterSettingsWidget::TEntryItem>::const_iterator iter;
	QListWidgetItem						*item;
	QVariant							variant;
	
	d_entries_list->clear();
	
	std::sort(d_entry_items.begin(), d_entry_items.end(), EntryItem_Sort_Function);
	
	for(iter = d_entry_items.begin();iter != d_entry_items.end();++iter)
	{
		item = new QListWidgetItem();
		variant.setValue((*iter));
		item->setData(Qt::UserRole,variant);
		
		item->setText(QString("Undulations: %1 Ph: %2 Amp: %3").arg((*iter).undulations).arg((*iter).phase,0,'f',2).arg((*iter).amplitude,0,'f',3));
		d_entries_list->addItem(item);
	}
}

bool TParameterSettingsWidget::Save(
	const QString 						&data_path) const
{
	std::vector<TParameterSettingsWidget::TEntryItem>::const_iterator iter;
	TXmlFile							xmlfile;
	QDomElement							en1;
	QDomElement							en2;

	try
	{
		xmlfile.Set_File_Name(QString("%1/parameters.xml").arg(data_path));
		xmlfile.Reset("Parameters",1);
				
		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Modulation"));
		xmlfile.Write_Text_Node(&en1,QString("%1").arg(this->Modulation()));
		
		for(iter = d_entry_items.begin();iter != d_entry_items.end();++iter)
		{
			en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Entry_Item"));
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Amplitude"));
			xmlfile.Write_Text_Node(&en2,QString("%1").arg((*iter).amplitude,0,'f',3));
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Phase"));
			xmlfile.Write_Text_Node(&en2,QString("%1").arg((*iter).phase,0,'f',3));

			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Undulations"));
			xmlfile.Write_Text_Node(&en2,QString("%1").arg((*iter).undulations));
		}

		xmlfile.Write_File();
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Saving parameters exception: %1").arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

void TParameterSettingsWidget::Load_Version1(
	TXmlFile							* const xml_file)
{
	QDomElement							en1;
	QDomElement							en2;
	TParameterSettingsWidget::TEntryItem entry_item;
	
	assert(xml_file);
	
	d_entry_items.clear();
			
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Modulation"));
	this->Set_Modulation(xml_file->Read_Text_Node(&en1).toInt());

	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Entry_Item"),true);
	while(!en1.isNull())
	{
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Amplitude"));
		entry_item.amplitude = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Phase"));
		entry_item.phase = xml_file->Read_Text_Node(&en2).toDouble();

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Undulations"));
		entry_item.undulations = xml_file->Read_Text_Node(&en2).toInt();
		
		d_entry_items.push_back(entry_item);
			
		en1 = xml_file->Get_Sibling_Node(&en1,QStringLiteral("Entry_Item"),true);
	}
}
